from typing import List

import rlp

from bxgateway.messages.eth.serializers.block_header import BlockHeader
from bxgateway.messages.eth.serializers.transaction import Transaction


# pyre-fixme[13]: Attribute `header` is never initialized.
# pyre-fixme[13]: Attribute `transactions` is never initialized.
# pyre-fixme[13]: Attribute `uncles` is never initialized.
class Block(rlp.Serializable):

    fields = [
        ("header", BlockHeader),
        ("transactions", rlp.sedes.CountableList(Transaction)),
        ("uncles", rlp.sedes.CountableList(BlockHeader))
    ]

    header: BlockHeader
    transactions: List[Transaction]
    uncles: List[BlockHeader]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Optional, List, Set, Union, NamedTuple

from bxcommon.messages.abstract_message import AbstractMessage

from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.models.quota_type_model import QuotaType
from bxcommon.utils.object_hash import Sha256Hash, convert
from bxcommon.utils.memory_utils import SpecialMemoryProperties, SpecialTuple

from bxgateway.utils.block_info import BlockInfo


class BlockDecompressionResult(NamedTuple):
    block_msg: Optional[AbstractMessage]
    block_info: BlockInfo
    unknown_short_ids: List[int]
    unknown_tx_hashes: List[Sha256Hash]


class AbstractMessageConverter(SpecialMemoryProperties, metaclass=ABCMeta):
    """
    Message converter abstract class.

    Converts messages of specific blockchain protocol to internal messages
    """

    @abstractmethod
    def tx_to_bx_txs(self, tx_msg, network_num: int, quota_type: Optional[QuotaType] = None) -> \
            List[Tuple[TxMessage, Sha256Hash, Union[bytearray, memoryview]]]:
        """
        Converts blockchain transactions message to internal transaction message

        :param tx_msg: blockchain transactions message
        :param network_num: blockchain network number
        :param quota_type: the quota type to assign to the BDN transaction.
        :return: array of tuples (transaction message, transaction hash, transaction bytes)
        """

        pass

    @abstractmethod
    def bx_tx_to_tx(self, bx_tx_msg):
        """
        Converts internal transaction message to blockchain transactions message

        :param bx_tx_msg: internal transaction message
        :return: blockchain transactions message
        """

        pass

    @abstractmethod
    def block_to_bx_block(self, block_msg, tx_service) -> Tuple[memoryview, BlockInfo]:
        """
        Convert blockchain block message to internal broadcast message with transactions replaced with short ids

        :param block_msg: blockchain new block message
        :param tx_service: Transactions service
        :return: Internal broadcast message bytes (bytearray), tuple (txs count, previous block hash, short ids)
        """

        pass

    @abstractmethod
    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        """
        Converts internal broadcast message to blockchain new block message

        Returns None for block message if any of the transactions shorts ids or hashes are unknown

        :param bx_block_msg: internal broadcast message bytes
        :param tx_service: Transactions service
        :return: block decompression result
        """

        pass

    @abstractmethod
    def bdn_tx_to_bx_tx(
            self,
            raw_tx: Union[bytes, bytearray, memoryview],
            network_num: int,
            quota_type: Optional[QuotaType] = None
    ) -> TxMessage:
        """
        Convert a raw transaction which arrived from an RPC request into bx transaction.
        :param raw_tx: The raw transaction bytes.
        :param network_num: the network number.
        :param quota_type: the quota type to assign to the BDN transaction.
        :return: bx transaction.
        """
        pass

    def encode_raw_msg(self, raw_msg: str) -> bytes:
        """
        Encode a raw message string into bytes
        :param raw_msg: the raw message to encode
        :return: binary encoded message
        :raise ValueError: if the encoding fails
        """
        # TODO: remove this method, access hex to bytes directly
        return convert.hex_to_bytes(raw_msg)

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        return super(AbstractMessageConverter, self).special_memory_size(ids)



from typing import Dict, Optional, List, Any

from bxgateway.feed.feed import Feed
from bxgateway.feed.subscriber import Subscriber
from bxutils import logging

logger = logging.get_logger(__name__)


class FeedManager:
    feeds: Dict[str, Feed]

    def __init__(self) -> None:
        self.feeds = {}

    def __contains__(self, item):
        return item in self.feeds

    def register_feed(self, feed: Feed) -> None:
        if feed.name in self.feeds:
            raise ValueError(
                f"Cannot register two feeds with the same name: {feed.name}"
            )

        self.feeds[feed.name] = feed

    def subscribe_to_feed(
        self, name: str, include_fields: Optional[List[str]] = None
    ) -> Optional[Subscriber]:
        if name in self.feeds:
            subscriber = self.feeds[name].subscribe(include_fields)
            logger.debug(
                "Creating new subscriber ({}) to {}",
                subscriber.subscription_id,
                name
            )
            return subscriber
        else:
            return None

    def unsubscribe_from_feed(
        self, name: str, subscriber_id: str
    ) -> Optional[Subscriber]:
        subscriber = self.feeds[name].unsubscribe(subscriber_id)
        if subscriber is not None:
            logger.debug(
                "Unsubscribing subscriber ({}) from {}",
                subscriber.subscription_id,
                name
            )
        return subscriber

    def publish_to_feed(self, name: str, message: Any) -> None:
        if name in self.feeds:
            self.feeds[name].publish(message)

    def get_feed_fields(self, feed_name: str) -> List[str]:
        return self.feeds[feed_name].FIELDS

    def any_subscribers(self) -> bool:
        return any(feed.subscriber_count() > 0 for feed in self.feeds.values())





from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.messages.eth.protocol.get_block_bodies_eth_protocol_message import (
    GetBlockBodiesEthProtocolMessage,
)
from bxgateway.messages.eth.protocol.get_block_headers_eth_protocol_message import (
    GetBlockHeadersEthProtocolMessage,
)
from bxgateway.services.block_processing_service import BlockProcessingService
from bxutils import logging

logger = logging.get_logger(__name__)


class EthBlockProcessingService(BlockProcessingService):
    _node: "bxgateway.connections.eth.eth_gateway_node.EthGatewayNode"

    def try_process_get_block_headers_request(
        self, msg: GetBlockHeadersEthProtocolMessage
    ) -> bool:

        block_queuing_service = self._node.block_queuing_service
        block_hash = msg.get_block_hash()

        if block_hash is not None:
            logger.trace(
                "Checking for headers by hash ({}) in local block cache...",
                block_hash,
            )
            (
                success, requested_block_hashes
            ) = block_queuing_service.get_block_hashes_starting_from_hash(
                block_hash,
                msg.get_amount(),
                msg.get_skip(),
                bool(msg.get_reverse()),
            )
        else:
            block_number = msg.get_block_number()
            if block_number:
                logger.trace(
                    "Checking for headers by block number ({}) "
                    "in local block cache",
                    block_number,
                )
                (
                    success, requested_block_hashes
                ) = block_queuing_service.get_block_hashes_starting_from_height(
                    block_number,
                    msg.get_amount(),
                    msg.get_skip(),
                    bool(msg.get_reverse()),
                )
            else:
                logger.debug(
                    "Unexpectedly, request for headers did not contain "
                    "block hash or block number. Skipping."
                )
                return False

        if success:
            return block_queuing_service.try_send_headers_to_node(
                requested_block_hashes
            )
        else:
            logger.trace(
                "Could not find requested block hashes. "
                "Forwarding to remote blockchain connection."
            )
            return False

    def try_process_get_block_bodies_request(
        self, msg: GetBlockBodiesEthProtocolMessage
    ) -> bool:
        block_hashes = msg.get_block_hashes()
        logger.trace("Checking for bodies in local block cache...")

        return self._node.block_queuing_service.try_send_bodies_to_node(
            block_hashes
        )



from collections import defaultdict
from typing import Dict, Optional


class PeerStatsMessage:
    bytes: int
    count: int

    def __init__(self):
        self.bytes = 0
        self.count = 0


class PeerStats:
    address: str
    peer_id: Optional[str]
    messages_received: Dict[str, PeerStatsMessage]
    messages_sent: PeerStatsMessage
    peer_total_received: int
    peer_total_sent: int
    ping_max: float

    def __init__(self):
        self.address = ""
        self.peer_id = None
        self.messages_received = defaultdict(PeerStatsMessage)
        self.messages_sent = PeerStatsMessage()
        self.peer_total_received = 0
        self.peer_total_sent = 0
        self.ping_max = 0



from typing import NamedTuple

from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.feed.feed import Feed


class RawTransactionFeedEntry:
    tx_hash: str
    tx_contents: str

    def __init__(
        self,
        tx_hash: Sha256Hash,
        tx_contents: memoryview
    ) -> None:
        self.tx_hash = str(tx_hash)
        self.tx_contents = convert.bytes_to_hex(tx_contents)

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, RawTransactionFeedEntry)
            and other.tx_hash == self.tx_hash
            and other.tx_contents == self.tx_contents
        )


class RawTransaction(NamedTuple):
    tx_hash: Sha256Hash
    tx_contents: memoryview


class NewTransactionFeed(Feed[RawTransactionFeedEntry, RawTransaction]):
    NAME = "newTxs"
    FIELDS = ["tx_hash", "tx_contents"]

    def __init__(self) -> None:
        super().__init__(self.NAME)

    def serialize(self, raw_message: RawTransaction) -> RawTransactionFeedEntry:
        return RawTransactionFeedEntry(raw_message.tx_hash, raw_message.tx_contents)



from typing import TYPE_CHECKING, Callable, List

from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc.rpc_errors import RpcInvalidParams, RpcInternalError, RpcAccountIdError
from bxgateway.feed.feed_manager import FeedManager
from bxgateway.feed.subscriber import Subscriber

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class SubscribeRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):
    help = {
        "params": "[feed_name, {\"include\": [field_1, field_2]}].\n"
                  "Available feeds: newTxs, pendingTxs\n"
                  "Available fields: tx_hash, tx_contents (default: all)",
        "description": "Subscribe to a named feed for notifications"
    }

    def __init__(
        self,
        request: BxJsonRpcRequest,
        node: "AbstractGatewayNode",
        feed_manager: FeedManager,
        subscribe_handler: Callable[[Subscriber, str], None],
    ) -> None:
        self.feed_name = ""
        self.include: List[str] = []
        self.feed_manager = feed_manager
        self.subscribe_handler = subscribe_handler

        super().__init__(request, node)

        assert self.feed_name != ""

    def validate_params(self) -> None:
        if not self.feed_manager.feeds:
            raise RpcAccountIdError(
                self.request_id,
                f"Account does not have access to the transaction streaming service."
            )
        
        params = self.params
        if not isinstance(params, list) or len(params) != 2:
            raise RpcInvalidParams(
                self.request_id,
                "Subscribe RPC request params must be a list of length 2."
            )

        feed_name, options = params
        feed = self.feed_manager.feeds.get(feed_name)
        if feed is None:
            raise RpcInvalidParams(
                self.request_id,
                f"{feed_name} is an invalid feed. "
                f"Available feeds: {list(self.feed_manager.feeds)}"
            )

        self.node.on_new_subscriber_request()

        self.feed_name = feed_name

        available_fields = self.feed_manager.get_feed_fields(feed_name)
        invalid_options = RpcInvalidParams(
            self.request_id,
            f"{options} is not a valid set of options. "
            "Valid format/fields: {\"include\": "f"{available_fields}""}."
        )
        if not isinstance(options, dict):
            raise invalid_options

        include = options.get("include", None)
        if include is not None:
            if not isinstance(include, list):
                raise invalid_options

            if any(
                included_field not in available_fields
                for included_field in include
            ):
                raise invalid_options

            self.include = include

    async def process_request(self) -> JsonRpcResponse:
        params = self.params
        assert isinstance(params, list)

        options = params[1]
        assert isinstance(options, dict)

        subscriber = self.feed_manager.subscribe_to_feed(
            self.feed_name, options.get("include", None)
        )
        assert subscriber is not None  # already validated
        self.subscribe_handler(subscriber, self.feed_name)

        return JsonRpcResponse(
            self.request_id,
            subscriber.subscription_id
        )





from typing import TYPE_CHECKING
import asyncio

from bxcommon.models.quota_type_model import QuotaType
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.exceptions import ParseError
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_blxr_transaction_rpc_request import \
    AbstractBlxrTransactionRpcRequest
from bxcommon.rpc.rpc_errors import RpcInvalidParams, RpcAccountIdError
from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.transaction_stat_event_type import TransactionStatEventType
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxutils import logging, log_messages as common_log_messages

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(__name__)


class GatewayBlxrTransactionRpcRequest(AbstractBlxrTransactionRpcRequest["AbstractGatewayNode"]):
    QUOTA_TYPE: str = "quota_type"
    SYNCHRONOUS = rpc_constants.SYNCHRONOUS_PARAMS_KEY

    def __init__(self, request: BxJsonRpcRequest, node: "AbstractGatewayNode") -> None:
        super().__init__(request, node)
        params = self.params
        assert isinstance(params, dict)

        lowercase_true = str(True).lower()
        self.synchronous = \
            params.get(self.SYNCHRONOUS, lowercase_true).lower() == lowercase_true

    async def process_request(self) -> JsonRpcResponse:
        params = self.params
        assert isinstance(params, dict)

        account_id = self.get_account_id()
        if not account_id:
            raise RpcAccountIdError(
                self.request_id,
                "Gateway does not have an associated account. Please register the gateway with an account to submit "
                "transactions through RPC."
            )

        transaction_str: str = params[rpc_constants.TRANSACTION_PARAMS_KEY]
        network_num = self.get_network_num()
        quota_type = QuotaType.PAID_DAILY_QUOTA
        return await self.process_transaction(network_num, account_id, quota_type, transaction_str)

    async def process_transaction(
        self, network_num: int, account_id: str, quota_type: QuotaType, transaction_str: str
    ) -> JsonRpcResponse:

        if not(self.node.account_model and self.node.account_model.cloud_api.is_service_valid()):
            raise RpcAccountIdError(
                self.request_id, "Account does not have cloud-api service. Please contact bloXroute support."
            )

        if self.synchronous:
            return await self.post_process_transaction(
                network_num, account_id, quota_type, transaction_str
            )
        else:
            asyncio.create_task(
                self.post_process_transaction(
                    network_num, account_id, quota_type, transaction_str
                )
            )
        return JsonRpcResponse(
            self.request_id,
            {
                "tx_hash": "not available with async",
                "quota_type": quota_type.name.lower(),
                "synchronous": str(self.synchronous)
            }
        )

    async def post_process_transaction(
        self, network_num: int, account_id: str, quota_type: QuotaType, transaction_str: str
    ) -> JsonRpcResponse:
        try:
            message_converter = self.node.message_converter
            assert message_converter is not None, "Invalid server state!"
            transaction = message_converter.encode_raw_msg(transaction_str)
            bx_tx = message_converter.bdn_tx_to_bx_tx(transaction, network_num, quota_type)
        except (ValueError, ParseError) as e:
            logger.error(common_log_messages.RPC_COULD_NOT_PARSE_TRANSACTION, e)
            raise RpcInvalidParams(
                self.request_id, f"Invalid transaction param: {transaction_str}"
            )
        tx_service = self.node.get_tx_service()
        tx_hash = bx_tx.tx_hash()
        if tx_service.has_transaction_contents(tx_hash) or tx_service.removed_transaction(tx_hash):
            short_id = tx_service.get_short_id(tx_hash)
            tx_stats.add_tx_by_hash_event(
                tx_hash,
                TransactionStatEventType.TX_RECEIVED_FROM_RPC_REQUEST_IGNORE_SEEN,
                network_num,
                account_id=account_id, short_id=short_id
            )
            tx_json = {
                "tx_hash": str(tx_hash),
                "quota_type": quota_type.name.lower(),
                "account_id": account_id,
            }
            return self.ok(tx_json)
        tx_stats.add_tx_by_hash_event(
            tx_hash,
            TransactionStatEventType.TX_RECEIVED_FROM_RPC_REQUEST,
            network_num,
            account_id=account_id
        )
        # All connections outside of this one is a bloXroute server
        broadcast_peers = self.node.broadcast(bx_tx, connection_types=[ConnectionType.RELAY_TRANSACTION])
        tx_stats.add_tx_by_hash_event(
            tx_hash,
            TransactionStatEventType.TX_SENT_FROM_GATEWAY_TO_PEERS,
            network_num,
            peers=stats_format.connections(broadcast_peers)
        )
        tx_stats.add_tx_by_hash_event(
            tx_hash,
            TransactionStatEventType.TX_GATEWAY_RPC_RESPONSE_SENT,
            network_num
        )
        tx_service.set_transaction_contents(tx_hash, bx_tx.tx_val())
        tx_json = {
            "tx_hash": str(tx_hash),
            "quota_type": quota_type.name.lower(),
            "account_id": account_id
        }
        return self.ok(tx_json)

    def get_network_num(self) -> int:
        return self.node.network_num

    def get_account_id(self) -> str:
        return self.node.account_id



from typing import Union, Optional, Dict, Callable, List

from bxcommon.exceptions import ParseError
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.models.blockchain_network_model import BlockchainNetworkModel
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.models.quota_type_model import QuotaType
from bxcommon.utils.blockchain_utils.btc import btc_common_util
from bxcommon.utils.blockchain_utils.eth import eth_common_util
from bxcommon.utils.blockchain_utils.ont import ont_common_util


protocol_to_bdn_tx_to_bx_tx: Dict[BlockchainProtocol, Callable[
    [Union[bytes, bytearray, memoryview], int, Optional[QuotaType]],
    TxMessage
]]
network_num_to_protocol: Dict[int, str]


def init(blockchain_networks: List[BlockchainNetworkModel]):
    # pylint: disable=global-statement
    global network_num_to_protocol
    network_num_to_protocol = {}
    for blockchain_network in blockchain_networks:
        network_num_to_protocol[blockchain_network.network_num] = blockchain_network.protocol.lower()


def bdn_tx_to_bx_tx(
        raw_tx: Union[bytes, bytearray, memoryview],
        network_num: int,
        quota_type: Optional[QuotaType] = None
) -> TxMessage:
    try:
        current_protocol = network_num_to_protocol[network_num]
    except KeyError:
        raise ValueError(f"{network_num} does not exist.")
    blockchain_protocol = BlockchainProtocol(current_protocol)
    return protocol_to_bdn_tx_to_bx_tx[blockchain_protocol](raw_tx, network_num, quota_type)


def btc_bdn_tx_to_bx_tx(
        raw_tx: Union[bytes, bytearray, memoryview],
        network_num: int,
        quota_type: Optional[QuotaType] = None
) -> TxMessage:
    if isinstance(raw_tx, bytes):
        raw_tx = bytearray(raw_tx)
    try:
        tx_hash = btc_common_util.get_txid(raw_tx)
    except IndexError:
        raise ValueError(f"Invalid raw transaction provided.")
    return TxMessage(
        message_hash=tx_hash, network_num=network_num, tx_val=raw_tx, quota_type=quota_type
    )


def eth_bdn_tx_to_bx_tx(
        raw_tx: Union[bytes, bytearray, memoryview],
        network_num: int,
        quota_type: Optional[QuotaType] = None
) -> TxMessage:
    if isinstance(raw_tx, bytes):
        raw_tx = bytearray(raw_tx)
    bx_tx, tx_item_length, tx_item_start = eth_common_util.raw_tx_to_bx_tx(raw_tx, 0, network_num, quota_type)
    if tx_item_length + tx_item_start != len(raw_tx):
        raise ParseError(raw_tx)
    return bx_tx


def ont_bdn_tx_to_bx_tx(
        raw_tx: Union[bytes, bytearray, memoryview],
        network_num: int,
        quota_type: Optional[QuotaType] = None
) -> TxMessage:
    if isinstance(raw_tx, bytes):
        raw_tx = bytearray(raw_tx)
    try:
        tx_hash, _ = ont_common_util.get_txid(raw_tx)
    except IndexError:
        raise ValueError(f"Invalid raw transaction provided.")
    return TxMessage(
        message_hash=tx_hash, network_num=network_num, tx_val=raw_tx, quota_type=quota_type
    )


protocol_to_bdn_tx_to_bx_tx = {
    BlockchainProtocol.BITCOIN: btc_bdn_tx_to_bx_tx,
    BlockchainProtocol.BITCOINCASH: btc_bdn_tx_to_bx_tx,
    BlockchainProtocol.ETHEREUM: eth_bdn_tx_to_bx_tx,
    BlockchainProtocol.ONTOLOGY: ont_bdn_tx_to_bx_tx

}



import typing
from asyncio import BaseTransport, Transport, BaseProtocol
from typing import TYPE_CHECKING, Optional

from cryptography.x509 import Certificate

from bxcommon import constants
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.network_direction import NetworkDirection
from bxcommon.network.socket_connection_state import SocketConnectionState
from bxutils import logging
from bxutils.logging import LogRecordType
from bxutils.ssl import ssl_certificate_factory

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode

logger = logging.get_logger(__name__)
network_troubleshooting_logger = logging.get_logger(LogRecordType.NetworkTroubleshooting, __name__)


class AbstractSocketConnectionProtocol(BaseProtocol):
    transport: Optional[Transport]
    file_no: int
    endpoint: IpEndpoint
    direction: NetworkDirection
    can_send: bool
    state: SocketConnectionState
    is_ssl: bool
    _node: "AbstractNode"
    _should_retry: bool
    _initial_bytes: Optional[int]  # bytes for message tracker to ignore for handshake
    _receive_buf: bytearray = bytearray(constants.RECV_BUFSIZE)

    def __init__(
        self,
        node: "AbstractNode",
        endpoint: Optional[IpEndpoint] = None,
        is_ssl: bool = True,
    ):
        self._node = node
        self.transport: Optional[Transport] = None
        self.file_no = -1
        # pyre-fixme[8]: Attribute has type `IpEndpoint`; used as
        #  `Optional[IpEndpoint]`.
        self.endpoint = endpoint
        if self.endpoint is None:
            self.direction = NetworkDirection.INBOUND
        else:
            self.direction = NetworkDirection.OUTBOUND
        self.can_send = False
        self.state = SocketConnectionState.CONNECTING
        self.is_ssl = is_ssl
        self._should_retry = self.direction == NetworkDirection.OUTBOUND
        self._initial_bytes = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <{self.endpoint}, {self.direction.name}>"

    def connection_made(self, transport: BaseTransport) -> None:
        self.transport = typing.cast(Transport, transport)
        self.file_no = transport.get_extra_info("socket").fileno()
        if self.direction == NetworkDirection.INBOUND:
            self.endpoint = IpEndpoint(
                *transport.get_extra_info("peername")
            )
            logger.debug("[{}] - accepted connection.", self)
        self._node.on_connection_added(self)
        self.state = SocketConnectionState.INITIALIZED
        self.can_send = True
        self.send()
        logger.debug("[{}] - connection established successfully.", self)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        mark_connection_for_close = (
            SocketConnectionState.MARK_FOR_CLOSE not in self.state
        )
        self.state |= SocketConnectionState.MARK_FOR_CLOSE
        if not self._should_retry:
            self.state |= SocketConnectionState.DO_NOT_RETRY
        if exc is not None:
            logger.info(
                "[{}] - lost connection due to an error: {}, closing connection, should_retry: {}.",
                self,
                exc,
                self._should_retry,
            )
        else:
            logger.debug(
                "[{}] - lost connection with peer, should_retry: {}.",
                self,
                self._should_retry,
            )
        self._node.on_connection_closed(self.file_no, mark_connection_for_close)

    def pause_writing(self) -> None:
        self.can_send = False
        logger.trace("[{}] - paused writing.", self)

    def resume_writing(self) -> None:
        self.can_send = True
        self.send()
        logger.trace("[{}] - resumed writing.", self)

    def send(self) -> None:
        total_bytes_sent = 0
        total_bytes_attempted_to_send = 0

        while self.is_sendable():
            data = self._node.get_bytes_to_send(self.file_no)
            if not data:
                break

            transport = self.transport
            assert transport is not None, "Connection is broken!"

            transport_buffer_size = transport.get_write_buffer_size()
            total_transport_buffer_size = transport_buffer_size + len(data)
            if self._initial_bytes is None:
                self._initial_bytes = transport_buffer_size

            transport.write(data)
            self._node.on_bytes_sent(self.file_no, len(data))
            total_bytes_attempted_to_send += len(data)

            bytes_sent = total_transport_buffer_size - transport.get_write_buffer_size()
            total_bytes_sent += bytes_sent

            initial_bytes = self._initial_bytes
            assert initial_bytes is not None
            remaining_initial_bytes = max(0, initial_bytes - bytes_sent)
            bytes_sent -= initial_bytes
            self._initial_bytes = remaining_initial_bytes

            self._node.on_bytes_written_to_socket(self.file_no, bytes_sent)

        if total_bytes_sent:
            logger.trace("[{}] - sent {} bytes", self, total_bytes_sent)

    def pause_reading(self) -> None:
        if self.is_alive():
            assert self.transport is not None, "Connection is broken!"
            self.state |= SocketConnectionState.HALT_RECEIVE
            logger.trace("[{}] - paused reading.", self)

    def resume_reading(self) -> None:
        if self.is_alive():
            assert self.transport is not None, "Connection is broken!"
            # pylint bug
            # pylint: disable=invalid-unary-operand-type
            self.state &= ~SocketConnectionState.HALT_RECEIVE
            logger.trace("[{}] - resumed writing.", self)

    def mark_for_close(self, should_retry: bool = True) -> None:
        if SocketConnectionState.MARK_FOR_CLOSE in self.state:
            return
        self.state |= SocketConnectionState.MARK_FOR_CLOSE
        self._should_retry = should_retry

        transport = self.transport
        assert transport is not None, "Connection is broken!"
        transport.close()
        logger.debug(
            "[{}] - marked for close, retrying: {}.", self, should_retry
        )

    def is_alive(self) -> bool:
        return SocketConnectionState.MARK_FOR_CLOSE not in self.state

    def is_receivable(self) -> bool:
        return (
            self.is_alive()
            and SocketConnectionState.HALT_RECEIVE not in self.state
        )

    def is_sendable(self) -> bool:
        return (
            self.is_alive()
            and SocketConnectionState.INITIALIZED in self.state
            and self.can_send
        )

    def get_peer_certificate(self) -> Certificate:
        assert self.transport is not None, "Connection is broken!"
        try:
            return ssl_certificate_factory.get_transport_cert(self.transport)
        except ValueError as e:
            raise TypeError("Socket is not SSL type!") from e

    def get_write_buffer_size(self) -> int:
        transport = self.transport
        assert transport is not None, "Connection is broken!"
        if transport.is_closing():
            return 0
        else:
            return transport.get_write_buffer_size()



PROTOCOL_VERSION = 10

# PROTOCOL_VERSION 10 (6/8/2020)
# change maximum transaction count in BDN performance stats message
#
# PROTOCOL_VERSION 9 (4/17/2020)
# add broadcast_type to broadcast messages
#
# PROTOCOL_VERSION 8 (2/5/2020)
# add timestamp to tx message types
#
# PROTOCOL_VERSION 7
# added TxQuotaFlag to transaction messages



import asyncio
import uuid
from asyncio import QueueFull
from typing import Generic, TypeVar, Optional, List, Union, Dict, Any

from bxgateway import gateway_constants

T = TypeVar("T")


class Subscriber(Generic[T]):
    """
    Subscriber object for asynchronous listening to a feed, with functionality
    to filter received messages on certain fields.

    Any object can published on a feed (and received by this subscriber), though
    users should be careful about ensuring that the object is serializable if
    the object should be received over the websocket or IPC subscriptions.

    If `include_fields` is None, the message will be passed onward with no
    changes.
    """

    include_fields: Optional[List[str]]
    subscription_id: str
    messages: 'asyncio.Queue[Union[T, Dict[str, Any]]]'

    def __init__(self, include_fields: Optional[List[str]] = None) -> None:
        self.include_fields = include_fields
        self.subscription_id = str(uuid.uuid4())
        self.messages = asyncio.Queue(
            gateway_constants.RPC_SUBSCRIBER_MAX_QUEUE_SIZE
        )

    async def receive(self) -> Union[T, Dict[str, Any]]:
        """
        Receives the next message in the queue.

        This function will block until a new message is posted to the queue.
        """
        message = await self.messages.get()
        return message

    def queue(self, message: T) -> None:
        """
        Queues up a message, releasing all receiving listeners.

        If too many messages are queued without a listener, this task
        will eventually fail and must be handled.
        """
        include_fields = self.include_fields
        if include_fields is not None:
            if isinstance(message, dict):
                filtered_message = {
                    key: message[key] for key in include_fields
                }
            else:
                filtered_message = {
                    key: getattr(message, key) for key in include_fields
                }
            self.messages.put_nowait(filtered_message)
        else:
            self.messages.put_nowait(message)



import json
import os
import traceback
from datetime import date, time, datetime
from enum import Enum
from inspect import istraceback
from typing import Union, Any, Iterable, Collection, Optional, Dict

from bxutils import logging

logger = logging.get_logger(__name__)

SPECIAL_ITERABLE_TYPES = (type(dict().values()), type(dict().keys()),)


def is_iterable_no_collection(o):
    return isinstance(o, SPECIAL_ITERABLE_TYPES) or \
           (isinstance(o, Iterable) and not isinstance(o, Collection))


def to_json(obj: Any, remove_nulls: bool = False) -> str:
    if remove_nulls:
        clean_dict = {}
        for key, value in obj.__dict__.items():
            if value:
                clean_dict[key] = value
        return json.dumps(clean_dict, cls=EnhancedJSONEncoder)

    return json.dumps(obj, cls=EnhancedJSONEncoder)


def to_dict(obj: Any) -> Dict[str, Any]:
    return EnhancedJSONEncoder().as_dict(obj)


def load_json_from_file(json_file_path: str) -> Optional[Union[list, dict]]:
    node_json = None

    if os.path.isfile(json_file_path):
        try:
            with open(json_file_path) as json_file:
                node_json = json.load(json_file)
        except ValueError as e:
            logger.debug("Failed to parse json: %s", e)
        except OSError as e:
            logger.debug("Failed trying to check for a json file: %s", e)
    else:
        raise ValueError("Could not locate json file: %s", json_file_path)

    return node_json


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if is_iterable_no_collection(o):
            o = list(o)
        elif isinstance(o, (bytearray, memoryview)):
            o = bytes(o)
        if isinstance(o, Enum):
            return str(o)
        if hasattr(o, "__dict__"):
            if isinstance(o.__dict__, dict):
                return o.__dict__
            else:
                return str(o)
        if isinstance(o, (date, datetime, time)):
            return o.isoformat()
        if isinstance(o, bytes):
            try:
                return o.decode("utf-8")
            except UnicodeDecodeError:
                return str(o)
        if hasattr(o, "hexdigest"):
            return o.hexdigest()
        if hasattr(o, "hex_string"):
            return o.hex_string()
        if istraceback(o):
            return "".join(traceback.format_tb(o)).strip()
        return o

    def _encode(self, obj):
        obj = self.default(obj)
        if isinstance(obj, dict):
            return {self.default(self._encode(k)): self._encode(v) for k, v in obj.items()}
        elif isinstance(obj, list) or isinstance(obj, set):
            return [self._encode(l) for l in obj]
        else:
            return obj

    def encode(self, o) -> str:
        return super(EnhancedJSONEncoder, self).encode(self._encode(o))

    def as_dict(self, obj) -> Dict[str, Any]:
        return self._encode(obj)



import weakref
from typing import TYPE_CHECKING

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.btc.btc_remote_connection_protocol import BtcRemoteConnectionProtocol

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class BtcRemoteConnection(AbstractGatewayBlockchainConnection["AbstractGatewayNode"]):
    CONNECTION_TYPE = ConnectionType.REMOTE_BLOCKCHAIN_NODE

    def __init__(self, socket_connection: AbstractSocketConnectionProtocol, node: "AbstractGatewayNode"):
        super(BtcRemoteConnection, self).__init__(socket_connection, node)
        self.connection_protocol = weakref.ref(BtcRemoteConnectionProtocol(self))



from dataclasses import dataclass
from typing import Optional
from datetime import datetime, date

from bxcommon.models.bdn_service_model_config_base import BdnServiceModelConfigBase
from bxcommon import constants


@dataclass
class BdnAccountModelBase:
    account_id: str
    logical_account_name: str
    certificate: str
    # TODO change expire_date to datetime type
    expire_date: str = constants.EPOCH_DATE
    tx_free: Optional[BdnServiceModelConfigBase] = None
    tx_paid: BdnServiceModelConfigBase = BdnServiceModelConfigBase()
    block_paid: BdnServiceModelConfigBase = BdnServiceModelConfigBase()
    cloud_api: BdnServiceModelConfigBase = BdnServiceModelConfigBase()
    new_transaction_streaming: BdnServiceModelConfigBase = BdnServiceModelConfigBase()
    blockchain_protocol: Optional[str] = None
    blockchain_network: Optional[str] = None

    def is_account_valid(self) -> bool:
        today = datetime.utcnow().date()
        try:
            expire_date = date.fromisoformat(self.expire_date)
        except (KeyError, ValueError):
            return False

        return expire_date >= today



import struct
from typing import List, cast

import task_pool_executor as tpe

from bxcommon import constants
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.services.extension_transaction_service import ExtensionTransactionService
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode
from bxgateway.connections.ont.ont_gateway_node import OntGatewayNode
from bxgateway.gateway_opts import GatewayOpts
from bxgateway.services.gateway_transaction_service import GatewayTransactionService, \
    ProcessTransactionMessageFromNodeResult


class ExtensionGatewayTransactionService(ExtensionTransactionService, GatewayTransactionService):

    def process_transactions_message_from_node(self, msg
                                               ) -> List[ProcessTransactionMessageFromNodeResult]:
        opts = cast(GatewayOpts, self.node.opts)
        msg_bytes = msg.rawbytes()

        if isinstance(self.node, OntGatewayNode) and opts.process_node_txs_in_extension:
            ext_processing_results = memoryview(self.proxy.process_gateway_transaction_from_node(
                BlockchainProtocol.ONTOLOGY.value,
                tpe.InputBytes(msg_bytes)
            ))
        elif isinstance(self.node, EthGatewayNode) and opts.process_node_txs_in_extension:
            ext_processing_results = memoryview(self.proxy.process_gateway_transaction_from_node(
                BlockchainProtocol.ETHEREUM.value,
                tpe.InputBytes(msg_bytes)
            ))
        else:
            return GatewayTransactionService.process_transactions_message_from_node(self, msg)

        result = []

        offset = 0

        txs_count, = struct.unpack_from("<I", ext_processing_results, offset)
        offset += constants.UL_INT_SIZE_IN_BYTES

        while offset < len(ext_processing_results):
            seen, = struct.unpack_from("<?", ext_processing_results, offset)
            offset += 1

            tx_hash = Sha256Hash(ext_processing_results[offset:offset + crypto.SHA256_HASH_LEN])
            offset += crypto.SHA256_HASH_LEN
            tx_offset, = struct.unpack_from("<I", ext_processing_results, offset)
            offset += constants.UL_INT_SIZE_IN_BYTES
            tx_len, = struct.unpack_from("<I", ext_processing_results, offset)
            offset += constants.UL_INT_SIZE_IN_BYTES

            tx_contents = msg_bytes[tx_offset:tx_offset + tx_len]

            result.append(ProcessTransactionMessageFromNodeResult(
                seen,
                tx_hash,
                tx_contents,
                TxMessage(message_hash=tx_hash, network_num=self.network_num, tx_val=tx_contents)
            ))

            if not seen:
                transaction_cache_key = self._tx_hash_to_cache_key(tx_hash)
                self.set_transaction_contents_base(
                    tx_hash,
                    transaction_cache_key,
                    tx_contents,
                    False,
                    0,
                    False
                )

        assert txs_count == len(result)

        return result





import os
import random
import typing
from abc import abstractmethod

from bxutils import logging
from bxutils.logging.log_level import LogLevel

from bxcommon.services.transaction_service import TransactionService
from bxcommon.test_utils import helpers
from bxcommon.utils import convert, crypto
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash

from bxgateway import btc_constants
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.messages.btc.inventory_btc_message import InventoryType
from bxgateway.services.btc.abstract_btc_block_cleanup_service import AbstractBtcBlockCleanupService
from bxgateway.testing.abstract_block_cleanup_service_test import AbstractBlockCleanupServiceTest

logger = logging.get_logger(__name__)


class AbstractBtcBlockCleanupServiceTest(AbstractBlockCleanupServiceTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        logger.setLevel(LogLevel.INFO)

    def _get_sample_block(self, file_path):
        root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(file_path))))
        with open(os.path.join(root_dir, "btc_sample_block.txt")) as sample_file:
            btc_block = sample_file.read().strip("\n")
        buf = bytearray(convert.hex_to_bytes(btc_block))
        parsed_block = BlockBtcMessage(buf=buf)
        return parsed_block

    def _test_mark_blocks_and_request_cleanup(self):
        marked_block = BtcObjectHash(binary=helpers.generate_bytearray(btc_constants.BTC_SHA_HASH_LEN))
        prev_block = BtcObjectHash(binary=helpers.generate_bytearray(btc_constants.BTC_SHA_HASH_LEN))
        tracked_blocks = []
        self.cleanup_service.on_new_block_received(marked_block, prev_block)
        self.transaction_service.track_seen_short_ids(marked_block, [])
        for _ in range(self.block_confirmations_count - 1):
            tracked_block = BtcObjectHash(binary=helpers.generate_bytearray(btc_constants.BTC_SHA_HASH_LEN))
            self.transaction_service.track_seen_short_ids(tracked_block, [])
            tracked_blocks.append(tracked_block)
        unmarked_block = BtcObjectHash(binary=helpers.generate_bytearray(btc_constants.BTC_SHA_HASH_LEN))
        self.assertIsNone(self.cleanup_service.last_confirmed_block)
        self.cleanup_service.mark_blocks_and_request_cleanup([marked_block, *tracked_blocks])
        self.assertEqual(marked_block, self.cleanup_service.last_confirmed_block)
        self.assertTrue(self.cleanup_service.is_marked_for_cleanup(marked_block))
        self.assertFalse(self.cleanup_service.is_marked_for_cleanup(unmarked_block))
        self.assertEqual(marked_block, self.cleanup_service.last_confirmed_block)
        msg = self.node.send_to_node_messages.pop(-1)
        self.assertEqual(1, msg.count())
        self.assertEqual((InventoryType.MSG_BLOCK, marked_block), next(iter(msg)))

    def _test_block_cleanup(self):
        block_msg = self._get_sample_block(self._get_file_path())
        transactions = block_msg.txns()[:]
        block_hash = typing.cast(BtcObjectHash, block_msg.block_hash())
        random.shuffle(transactions)
        short_len = int(len(transactions) * 0.9)
        transactions_short = transactions[:short_len]
        unknown_transactions = transactions[short_len:]
        transaction_hashes = []
        for idx, tx in enumerate(transactions_short):
            tx_hash = BtcObjectHash(
                buf=crypto.double_sha256(tx),
                length=btc_constants.BTC_SHA_HASH_LEN
            )
            transaction_hashes.append(tx_hash)
            self.transaction_service.set_transaction_contents(tx_hash, tx)
            self.transaction_service.assign_short_id(tx_hash, idx + 1)
        for idx, tx in enumerate(unknown_transactions):
            tx_hash = BtcObjectHash(
                buf=crypto.double_sha256(tx),
                length=btc_constants.BTC_SHA_HASH_LEN
            )
            transaction_hashes.append(tx_hash)
            if idx % 2 == 0:
                self.transaction_service.set_transaction_contents(tx_hash, tx)
        self.cleanup_service._block_hash_marked_for_cleanup.add(block_hash)
        self.cleanup_service.clean_block_transactions(block_msg, self.transaction_service)
        self.assertEqual(0, self.transaction_service._total_tx_contents_size)
        for tx_hash in transaction_hashes:
            self.assertFalse(self.transaction_service.has_transaction_contents(tx_hash))

    @abstractmethod
    def _get_transaction_service(self) -> TransactionService:
        pass

    @abstractmethod
    def _get_cleanup_service(self) -> AbstractBtcBlockCleanupService:
        pass

    @abstractmethod
    def _get_file_path(self) -> str:
        pass



from abc import abstractmethod
from typing import Generic

from bxcommon.models.quota_type_model import QuotaType

from bxcommon.rpc import rpc_constants
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest, Node
from bxcommon.rpc.rpc_errors import RpcInvalidParams, RpcAccountIdError


class AbstractBlxrTransactionRpcRequest(AbstractRpcRequest, Generic[Node]):
    SYNCHRONOUS = rpc_constants.SYNCHRONOUS_PARAMS_KEY
    help = {
        "params": f"[Required - {rpc_constants.TRANSACTION_PARAMS_KEY}: [transaction payload in hex string format]]. "
                  f"Optional - {SYNCHRONOUS}: [True (wait for response from the relay - default), "
                  "False (don't wait for response)].",
        "description": "send transaction to the bloXroute BDN"
    }

    def validate_params(self) -> None:
        params = self.params
        if params is None or not isinstance(params, dict):
            raise RpcInvalidParams(
                self.request_id,
                "Params request field is either missing or not a dictionary type."
            )
        if rpc_constants.TRANSACTION_PARAMS_KEY not in params:
            raise RpcInvalidParams(
                self.request_id,
                f"Invalid transaction request params type: {self.params}"
            )

    async def process_request(self) -> JsonRpcResponse:
        params = self.params
        assert isinstance(params, dict)

        account_id = self.get_account_id()
        if not account_id:
            raise RpcAccountIdError(
                self.request_id,
                "Account ID is missing."
            )

        transaction_str: str = params[rpc_constants.TRANSACTION_PARAMS_KEY]
        network_num = self.get_network_num()
        quota_type = QuotaType.PAID_DAILY_QUOTA
        return await self.process_transaction(network_num, account_id, quota_type, transaction_str)

    @abstractmethod
    async def process_transaction(
        self, network_num: int, account_id: str, quota_type: QuotaType, transaction_str: str
    ) -> JsonRpcResponse:
        pass

    @abstractmethod
    def get_network_num(self) -> int:
        pass

    @abstractmethod
    def get_account_id(self) -> str:
        pass



import asyncio
import base64
from abc import abstractmethod
from asyncio import Future
from typing import Callable, Awaitable, Optional, TYPE_CHECKING, TypeVar, Generic
from aiohttp import web
from aiohttp.abc import StreamResponse
from aiohttp.web import Application, Request, Response, AppRunner, TCPSite
from aiohttp.web_exceptions import HTTPClientError, HTTPBadRequest, HTTPInternalServerError
from aiohttp.web_exceptions import HTTPUnauthorized

from bxcommon.rpc import rpc_constants
from bxcommon.rpc.https.http_rpc_handler import HttpRpcHandler
from bxcommon.rpc.https.request_formatter import RequestFormatter
from bxcommon.rpc.https.response_formatter import ResponseFormatter
from bxcommon.rpc.json_rpc_response import JsonRpcResponse

from bxcommon.rpc.rpc_errors import RpcError, RpcParseError, RpcMethodNotFound, \
    RpcInvalidParams, RpcAccountIdError
from bxutils.encoding import json_encoder

from bxutils import logging

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode

logger = logging.get_logger(__name__)
Node = TypeVar("Node", bound="AbstractNode")


@web.middleware
async def request_middleware(
    request: Request, handler: Callable[[Request], Awaitable[StreamResponse]]
) -> StreamResponse:
    request_formatter = RequestFormatter(request)
    logger.trace("Handling RPC request: {}.", request_formatter)
    response = await handler(request)
    logger.trace(
        "Finished handling request: {}, returning response: {}.",
        request_formatter,
        ResponseFormatter(response),
    )
    return response


class AbstractHttpRpcServer(Generic[Node]):
    RUN_SLEEP_INTERVAL_S: int = 5

    node: Node
    _app: Application
    _runner: AppRunner
    _site: Optional[TCPSite]
    _handler: HttpRpcHandler[Node]
    _stop_requested: bool
    _stop_waiter: Future
    _started: bool
    _encoded_auth: Optional[str]

    def __init__(self, node: Node) -> None:
        self.node = node
        self._app = Application(middlewares=[request_middleware])
        self._app.add_routes(
            [web.post("/", self.handle_request), web.get("/", self.handle_get_request)]
        )
        self._runner = AppRunner(self._app)
        self._site = None
        self._handler = self.request_handler()
        self._stop_requested = False
        self._stop_waiter = asyncio.get_event_loop().create_future()
        self._started = False
        self._encoded_auth = None
        rpc_user = self.node.opts.rpc_user
        if rpc_user:
            rpc_password = self.node.opts.rpc_password
            self._encoded_auth = base64.b64encode(
                f"{rpc_user}:{rpc_password}".encode("utf-8")
            ).decode("utf-8")

    @abstractmethod
    def authenticate_request(self, request: Request) -> None:
        pass

    @abstractmethod
    def request_handler(self) -> HttpRpcHandler:
        pass

    def status(self) -> bool:
        return self._started

    async def run(self) -> None:
        try:
            await self._start()
            while not self._stop_requested:
                await asyncio.sleep(self.RUN_SLEEP_INTERVAL_S)
        finally:
            self._stop_waiter.set_result(True)

    async def start(self) -> None:
        if self._started:
            return
        try:
            await self._start()
        finally:
            self._stop_waiter.set_result(True)

    async def stop(self) -> None:
        self._stop_requested = True
        await self._stop_waiter
        await self._runner.cleanup()
        self._started = False

    async def handle_request(self, request: Request) -> Response:
        try:
            self.authenticate_request(request)
            return await self._handler.handle_request(request)
        except HTTPClientError as e:
            return self._format_http_error(e)
        except RpcAccountIdError as e:
            return self._format_rpc_error(e, HTTPUnauthorized.status_code)
        except (RpcParseError, RpcMethodNotFound, RpcInvalidParams) as e:
            return self._format_rpc_error(e, HTTPBadRequest.status_code)
        except RpcError as e:
            return self._format_rpc_error(e, HTTPInternalServerError.status_code)

    async def handle_get_request(self, request: Request) -> Response:
        try:
            self.authenticate_request(request)
        except HTTPUnauthorized as e:
            return self._format_http_error(e)
        else:
            response_dict = {
                "result": {
                    "required_request_type": "POST",
                    "required_headers": [
                        {
                            rpc_constants.CONTENT_TYPE_HEADER_KEY: rpc_constants.PLAIN_HEADER_TYPE
                        }
                    ],
                    "payload_structures": await self._handler.help(),
                }
            }
            return web.json_response(json_encoder.to_json(response_dict))

    async def _start(self) -> None:
        self._started = True
        await self._runner.setup()
        opts = self.node.opts

        # TODO: add ssl certificate
        site = TCPSite(self._runner, opts.rpc_host, opts.rpc_port)
        self._site = site
        await site.start()

    def _format_http_error(self, client_error: HTTPClientError) -> HTTPClientError:
        err_msg = client_error.text
        code = client_error.status_code
        response_json = {
            "result": None,
            "error": err_msg,
            "code": code,
            "message": err_msg,
        }
        client_error.text = json_encoder.to_json(response_json)
        return client_error

    def _format_rpc_error(self, rpc_error: RpcError, status_code: int) -> Response:
        return web.json_response(
            JsonRpcResponse(rpc_error.id, error=rpc_error).to_jsons(),
            status=status_code
        )



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, List, Union

from bxcommon.connections.connection_type import ConnectionType
from bxgateway.utils.logging.status import summary
from bxgateway.utils.logging.status.blockchain_connection import BlockchainConnection
from bxgateway.utils.logging.status.connection_state import ConnectionState
from bxgateway.utils.logging.status.gateway_status import GatewayStatus
from bxgateway.utils.logging.status.relay_connection import RelayConnection
from bxgateway.utils.logging.status.summary import Summary


@dataclass
class Network:
    block_relays: List[RelayConnection]
    transaction_relays: List[RelayConnection]
    blockchain_nodes: List[BlockchainConnection]
    remote_blockchain_nodes: List[BlockchainConnection]

    def get_summary(self, ip_address: str, continent: str, country: str, update_required: bool,
                    account_id: Optional[str], quota_level: Optional[int]) -> Summary:
        block_relay_connections_state = _connections_states_info(self.block_relays)
        transaction_relay_connections_state = _connections_states_info(self.transaction_relays)
        blockchain_node_connections_state = _connections_states_info(self.blockchain_nodes)
        remote_blockchain_node_connections_state = _connections_states_info(self.remote_blockchain_nodes)
        gateway_status = self._get_gateway_status()

        return Summary(gateway_status, summary.gateway_status_get_account_info(account_id),
                       block_relay_connections_state, transaction_relay_connections_state,
                       blockchain_node_connections_state, remote_blockchain_node_connections_state,
                       ip_address, continent, country, update_required,
                       summary.gateway_status_get_quota_level(quota_level))

    def clear(self):
        self.block_relays = []
        self.transaction_relays = []
        self.blockchain_nodes = []
        self.remote_blockchain_nodes = []

    def add_connection(self, conn: ConnectionType, desc: Optional[str] = None, file_no: Optional[str] = None,
                       peer_id: Optional[str] = None) -> None:
        ip_addr = None if desc is None else desc.split()[0]
        port = None if desc is None else desc.split()[1]
        current_time = _get_current_time()

        if conn == ConnectionType.RELAY_BLOCK:
            self.block_relays.append(RelayConnection(ip_addr, port, file_no, peer_id, current_time))
        elif conn == ConnectionType.RELAY_TRANSACTION:
            self.transaction_relays.append(RelayConnection(ip_addr, port, file_no, peer_id, current_time))
        elif conn == ConnectionType.BLOCKCHAIN_NODE:
            self.blockchain_nodes.append(BlockchainConnection(ip_addr, port, file_no, current_time))
        else:
            self.remote_blockchain_nodes.append(BlockchainConnection(ip_addr, port, file_no, current_time))

    def _get_gateway_status(self) -> GatewayStatus:
        return GatewayStatus.ONLINE if _check_connections_established(
            self.block_relays) and _check_connections_established(
            self.transaction_relays) and _check_connections_established(
            self.blockchain_nodes) and _check_connections_established(
            self.remote_blockchain_nodes) else GatewayStatus.WITH_ERRORS


def _get_current_time() -> str:
    return "UTC " + str(datetime.utcnow())


def _check_connections_established(connections: Union[List[RelayConnection], List[BlockchainConnection]]) -> bool:
    return len(connections) > 0 and all([conn.get_connection_state() == ConnectionState.ESTABLISHED for conn in connections])


def _connections_states_info(connections: Union[List[RelayConnection], List[BlockchainConnection]]) -> ConnectionState:
    return ConnectionState.ESTABLISHED if _check_connections_established(connections) else ConnectionState.DISCONNECTED



from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.services.broadcast_service import BroadcastService


class GatewayBroadcastService(BroadcastService[AbstractMessage, AbstractConnection]):
    def should_broadcast_to_connection(self, message: AbstractMessage, connection: AbstractConnection) -> bool:
        # gateway does not really care about network numbers
        return True



import time
from datetime import datetime
from typing import Any, List, Union, Optional

import task_pool_executor as tpe

from bxcommon import constants
from bxcommon.services.transaction_service import TransactionService, TransactionCacheKeyType, \
    TransactionFromBdnGatewayProcessingResult
from bxcommon.services.transaction_service import TxRemovalReason
from bxcommon.utils import memory_utils
from bxcommon.utils.object_encoder import ObjectEncoder
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.proxy import task_pool_proxy
from bxcommon.utils.proxy.default_map_proxy import DefaultMapProxy
from bxcommon.utils.proxy.map_proxy import MapProxy
from bxcommon.utils.stats import hooks
from bxcommon.utils.stats.transaction_stat_event_type import TransactionStatEventType
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxutils import logging
from bxutils.logging import log_config
from bxutils.logging.log_record_type import LogRecordType

logger = logging.get_logger(__name__)
logger_memory_cleanup = logging.get_logger(LogRecordType.BlockCleanup, __name__)


class ExtensionTransactionService(TransactionService):

    def __init__(self, node, network_num):
        super(ExtensionTransactionService, self).__init__(node, network_num)

        # Log levels need to be set again to include the loggers created in this conditionally imported class
        log_config.lazy_set_log_level(node.opts.log_level_overrides)

        self.proxy = tpe.TransactionService(
            task_pool_proxy.get_pool_size(),
            node.opts.tx_mem_pool_bucket_size,
            self._get_final_tx_confirmations_count()
        )
        raw_encoder = ObjectEncoder.raw_encoder()
        self._tx_cache_key_to_short_ids = DefaultMapProxy(
            self.proxy.tx_hash_to_short_ids(), raw_encoder, raw_encoder
        )
        self._short_id_to_tx_cache_key = MapProxy(
            self.proxy.short_id_to_tx_hash(), raw_encoder, raw_encoder
        )
        content_encoder = ObjectEncoder(memoryview, tpe.InputBytes)
        self._tx_cache_key_to_contents = MapProxy(
            self.proxy.tx_hash_to_contents(), raw_encoder, content_encoder
        )
        self._tx_not_seen_in_blocks = self.proxy.tx_not_seen_in_blocks()

        self._tx_hash_to_time_removed = MapProxy(
            self.proxy.tx_hash_to_time_removed(), raw_encoder, raw_encoder
        )

    def track_seen_short_ids(self, block_hash, short_ids: List[int]) -> None:
        start_datetime = datetime.now()
        super(ExtensionTransactionService, self).track_seen_short_ids(block_hash, short_ids)
        wrapped_block_hash = tpe.Sha256(tpe.InputBytes(self._wrap_sha256(block_hash).binary))
        proxy_start_datetime = datetime.now()
        # TODO when refactoring add `block_hash` to proxy.track_seen_short_ids as
        # first parameter and change ds type in cpp
        result = self.proxy.track_seen_short_ids(wrapped_block_hash, tpe.UIntList(short_ids))
        removed_contents_size, dup_sids = result
        self.update_removed_transactions(removed_contents_size, dup_sids)
        logger_memory_cleanup.statistics(
            {
                "type": "MemoryCleanup",
                "event": "ExtensionTransactionServiceTrackSeenSummary",
                "seen_short_ids_count": len(short_ids),
                "total_content_size_removed": removed_contents_size,
                "total_duplicate_short_ids": len(dup_sids),
                "proxy_call_datetime": proxy_start_datetime,
                "data": self.get_cache_state_json(),
                "start_datetime": start_datetime,
                "block_hash": repr(block_hash)
            }
        )

    def set_final_tx_confirmations_count(self, val: int):
        super(ExtensionTransactionService, self).set_final_tx_confirmations_count(val)
        self.proxy.set_final_tx_confirmations_count(val)

    def on_block_cleaned_up(self, block_hash: Sha256Hash) -> None:
        super(ExtensionTransactionService, self).on_block_cleaned_up(block_hash)
        wrapped_block_hash = tpe.Sha256(tpe.InputBytes(block_hash.binary))
        self.proxy.on_block_cleaned_up(wrapped_block_hash)

    def get_tx_service_sync_buffer(self, sync_tx_content: bool) -> memoryview:
        byte_array_obj = self.proxy.get_tx_sync_buffer(self._total_tx_contents_size,
                                                       sync_tx_content)
        return memoryview(byte_array_obj)

    def update_removed_transactions(self, removed_content_size: int, short_ids: List[int]) -> None:
        self._total_tx_contents_size -= removed_content_size
        for short_id in short_ids:
            tx_stats.add_tx_by_hash_event(
                constants.UNKNOWN_TRANSACTION_HASH, TransactionStatEventType.TX_REMOVED_FROM_MEMORY,
                self.network_num, short_id, reason=TxRemovalReason.EXTENSION_BLOCK_CLEANUP.value
            )
            self._tx_assignment_expire_queue.remove(short_id)
            if self.node.opts.dump_removed_short_ids:
                self._removed_short_ids.add(short_id)

    def assign_short_id(self, transaction_hash: Sha256Hash, short_id: int):
        """
        Adds short id mapping for transaction and schedules an alarm to cleanup entry on expiration.
        :param transaction_hash: transaction long hash
        :param short_id: short id to be mapped to transaction
        """
        logger.trace("Assigning sid {} to transaction {}", short_id, transaction_hash)
        tx_cache_key = self._tx_hash_to_cache_key(transaction_hash)
        has_contents = self.proxy.assign_short_id(tx_cache_key, short_id)
        self.assign_short_id_base(transaction_hash, tx_cache_key, short_id, has_contents, False)

    def set_transaction_contents(
        self, transaction_hash: Sha256Hash, transaction_contents: Union[bytearray, memoryview],
        transaction_cache_key: Optional[TransactionCacheKeyType] = None
    ):
        """
        Adds transaction contents to transaction service cache with lookup key by transaction hash

        :param transaction_hash: transaction hash
        :param transaction_contents: transaction contents bytes
        :param transaction_cache_key: transaction cache key optional
        """
        if not transaction_cache_key:
            transaction_cache_key = self._tx_hash_to_cache_key(transaction_hash)

        assert isinstance(transaction_cache_key, tpe.Sha256)
        has_short_id, previous_size = self.proxy.set_transaction_contents(
            transaction_cache_key,
            tpe.InputBytes(transaction_contents))

        self.set_transaction_contents_base(
            transaction_hash,
            transaction_cache_key,
            transaction_contents,
            has_short_id,
            previous_size,
            False
        )

    def process_gateway_transaction_from_bdn(
        self,
        transaction_hash: Sha256Hash,
        short_id: int,
        transaction_contents: Union[bytearray, memoryview],
        is_compact: bool
    ) -> TransactionFromBdnGatewayProcessingResult:

        transaction_cache_key = self._tx_hash_to_cache_key(transaction_hash)

        ext_result = self.proxy.process_gateway_transaction_from_bdn(
            transaction_cache_key,
            tpe.InputBytes(transaction_contents),
            short_id,
            is_compact
        )

        result = TransactionFromBdnGatewayProcessingResult(
            ext_result.get_ignore_seen(),
            ext_result.get_existing_short_id(),
            ext_result.get_assigned_short_id(),
            ext_result.get_existing_contents(),
            ext_result.get_set_contents()
        )

        if result.set_content:
            has_short_id, previous_size = ext_result.get_set_contents_result()
            self.set_transaction_contents_base(
                transaction_hash,
                transaction_cache_key,
                transaction_contents,
                has_short_id,
                previous_size,
                False
            )

        if result.assigned_short_id:
            has_contents = result.existing_contents or result.set_content
            self.assign_short_id_base(
                transaction_hash,
                transaction_cache_key,
                short_id,
                has_contents,
                False
            )

        return result

    def log_tx_service_mem_stats(self):
        super(ExtensionTransactionService, self).log_tx_service_mem_stats()
        if self.node.opts.stats_calculate_actual_size:
            size_type = memory_utils.SizeType.OBJECT
        else:
            size_type = memory_utils.SizeType.ESTIMATE
        hooks.add_obj_mem_stats(
            self.__class__.__name__,
            self.network_num,
            self._tx_not_seen_in_blocks,
            "tx_not_seen_in_blocks",
            self.get_collection_mem_stats(
                self._tx_not_seen_in_blocks),
            object_item_count=len(self._tx_not_seen_in_blocks),
            object_type=memory_utils.ObjectType.BASE,
            size_type=size_type
        )

    def get_collection_mem_stats(self, collection_obj: Any,
                                 estimated_size: int = 0) -> memory_utils.ObjectSize:
        if self.get_object_type(collection_obj) == memory_utils.ObjectType.DEFAULT_MAP_PROXY:
            collection_size = collection_obj.map_obj.get_bytes_length()
            if collection_obj is self._tx_cache_key_to_short_ids:
                collection_size += (
                    len(self._short_id_to_tx_cache_key) * constants.UL_INT_SIZE_IN_BYTES)
            return memory_utils.ObjectSize(size=collection_size, flat_size=0, is_actual_size=True)
        else:
            return super(ExtensionTransactionService, self).get_collection_mem_stats(collection_obj,
                                                                                     estimated_size)

    def get_object_type(self, collection_obj: Any):
        super(ExtensionTransactionService, self).get_object_type(collection_obj)
        if isinstance(collection_obj, DefaultMapProxy):
            return memory_utils.ObjectType.DEFAULT_MAP_PROXY
        elif isinstance(collection_obj, MapProxy):
            return memory_utils.ObjectType.MAP_PROXY
        else:
            return memory_utils.ObjectType.BASE

    def _tx_hash_to_cache_key(self, transaction_hash) -> tpe.Sha256:  # pyre-ignore
        if isinstance(transaction_hash, Sha256Hash):
            return tpe.Sha256(tpe.InputBytes(transaction_hash.binary))

        if isinstance(transaction_hash, (bytes, bytearray, memoryview)):
            return tpe.Sha256(tpe.InputBytes(transaction_hash))

        if isinstance(transaction_hash, tpe.Sha256):
            return transaction_hash

        raise ValueError("Attempted to find cache entry with incorrect key type")

        # return transaction_hash

    def _tx_cache_key_to_hash(self, transaction_cache_key) -> Sha256Hash:
        if isinstance(transaction_cache_key, Sha256Hash):
            return transaction_cache_key

        if isinstance(transaction_cache_key, (bytes, bytearray, memoryview)):
            return Sha256Hash(transaction_cache_key)

        return Sha256Hash(bytearray(transaction_cache_key.binary()))

    def _track_seen_transaction(self, transaction_cache_key):
        super(ExtensionTransactionService, self)._track_seen_transaction(transaction_cache_key)
        self.proxy.track_seen_transaction(transaction_cache_key)

    def remove_transaction_by_short_id(self, short_id: int, remove_related_short_ids: bool = False,
                                       force: bool = False,
                                       removal_reason: TxRemovalReason = TxRemovalReason.UNKNOWN):
        # overriding this in order to handle removes triggered by either the mem limit or expiration queue
        # if the remove_related_short_ids is True than we assume the call originated by the track seen call
        # else we assume it was triggered by the cleanup.
        # this is only a temporary fix and the whole class hierarchy requires some refactoring!
        if remove_related_short_ids:
            self._tx_assignment_expire_queue.remove(short_id)
            tx_stats.add_tx_by_hash_event(
                constants.UNKNOWN_TRANSACTION_HASH, TransactionStatEventType.TX_REMOVED_FROM_MEMORY,
                self.network_num, short_id, reason=removal_reason.value
            )
            if self.node.opts.dump_removed_short_ids:
                self._removed_short_ids.add(short_id)
        else:
            super(ExtensionTransactionService, self).remove_transaction_by_short_id(short_id,
                                                                                    force=force,
                                                                                    removal_reason=removal_reason)

    def _clear(self):
        super(ExtensionTransactionService, self)._clear()
        self.proxy.clear_short_ids_seen_in_block()

    def _cleanup_removed_transactions_history(self):
        logger.trace(
            "Starting to cleanup transaction cache history for network "
            "number: {}.",
            self.network_num
        )
        history_len_before = len(self._tx_hash_to_time_removed)
        self._tx_hash_to_time_removed.map_obj.cleanup_removed_hashes_history(
            time.time(), self._removed_txs_expiration_time_s, constants.REMOVED_TRANSACTIONS_HISTORY_LENGTH_LIMIT
        )
        history_len_after = len(self._tx_hash_to_time_removed)
        logger.trace(
            "Finished cleanup transaction cache history. Size before: {}. "
            "Size after: {}.",
            history_len_before,
            history_len_after
        )

        return constants.REMOVED_TRANSACTIONS_HISTORY_CLEANUP_INTERVAL_S



from bxcommon import constants
from bxcommon.messages.bloxroute.bloxroute_message_factory import bloxroute_message_factory
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.protocol_version import PROTOCOL_VERSION
from bxcommon.messages.bloxroute.v6.bloxroute_message_factory_v6 import bloxroute_message_factory_v6
from bxcommon.messages.bloxroute.v6.message_converter_factory_v6 import message_converter_factory_v6
from bxcommon.messages.bloxroute.v7.bloxroute_message_factory_v7 import bloxroute_message_factory_v7
from bxcommon.messages.bloxroute.v7.message_converter_factory_v7 import message_converter_factory_v7
from bxcommon.messages.bloxroute.v8.bloxroute_message_factory_v8 import bloxroute_message_factory_v8
from bxcommon.messages.bloxroute.v8.message_converter_factory_v8 import message_converter_factory_v8
from bxcommon.messages.bloxroute.v9.bloxroute_message_factory_v9 import bloxroute_message_factory_v9
from bxcommon.messages.bloxroute.v9.message_converter_factory_v9 import message_converter_factory_v9
from bxcommon.messages.versioning.abstract_version_manager import AbstractVersionManager


class _BloxrouteVersionManager(AbstractVersionManager):
    CURRENT_PROTOCOL_VERSION = PROTOCOL_VERSION
    MIN_SUPPORTED_PROTOCOL_VERSION = 6
    VERSION_MESSAGE_MAIN_LENGTH = constants.VERSIONED_HELLO_MSG_MIN_PAYLOAD_LEN
    _PROTOCOL_TO_CONVERTER_FACTORY_MAPPING = {
        6: message_converter_factory_v6,
        7: message_converter_factory_v7,
        8: message_converter_factory_v8,
        9: message_converter_factory_v9,
    }
    _PROTOCOL_TO_FACTORY_MAPPING = {
        6: bloxroute_message_factory_v6,
        7: bloxroute_message_factory_v7,
        8: bloxroute_message_factory_v8,
        9: bloxroute_message_factory_v9,
        10: bloxroute_message_factory,
    }

    def __init__(self):
        super(_BloxrouteVersionManager, self).__init__()
        self.protocol_to_factory_mapping = self._PROTOCOL_TO_FACTORY_MAPPING
        self.protocol_to_converter_factory_mapping = self._PROTOCOL_TO_CONVERTER_FACTORY_MAPPING
        self.version_message_command = BloxrouteMessageType.HELLO


bloxroute_version_manager = _BloxrouteVersionManager()



import os
import random
import struct
import sys
import pyelliptic
import rlp
from Crypto.Cipher import AES
from rlp import sedes

from bxutils import logging

from bxcommon.exceptions import ParseError

from bxgateway import eth_constants
from bxgateway.eth_exceptions import InvalidKeyError, AuthenticationError, CipherNotInitializedError
from bxgateway.utils.eth import crypto_utils, rlp_utils
from bxgateway.utils.eth.eccx import ECCx
from bxgateway.utils.eth.frame import Frame

logger = logging.get_logger(__name__)


class RLPxCipher(object):

    def __init__(self, is_initiator, private_key, remote_public_key=None):

        self._ecc = ECCx(raw_private_key=private_key)

        if remote_public_key and not self._ecc.is_valid_key(remote_public_key):
            raise InvalidKeyError("Invalid remote pubkey")

        self._ephemeral_ecc = ECCx()

        self._is_initiator = is_initiator
        self._private_key = private_key
        self._remote_pubkey = remote_public_key

        self._is_eip8_auth = False
        self._remote_ephemeral_pubkey = None
        self._initiator_nonce = None
        self._responder_nonce = None
        self._auth_init = None
        self._auth_ack = None
        self._aes_secret = None
        self._token = None
        self._aes_enc = None
        self._aes_dec = None
        self._mac_enc = None
        self._egress_mac = None
        self._ingress_mac = None
        self._is_ready = False
        self._remote_version = 0

    def get_private_key(self):
        """
        Get private key of current ECCx object
        :return: private key
        """

        return self._private_key

    def is_ready(self):
        """
        Returns flag indicating if cipher is ready for encrypted messages exchange
        :return: True if cipher is ready, False otherwise
        """

        return self._is_ready

    def create_auth_message(self):
        """
        Generates authentication message bytes

        1. initiator generates ecdhe-random and nonce and creates auth
        2. initiator connects to remote and sends auth

        New:
        E(remote-pubk,
            S(ephemeral-privk, ecdh-shared-secret ^ nonce) ||
            H(ephemeral-pubk) || pubk || nonce || 0x0
        )
        Known:
        E(remote-pubk,
            S(ephemeral-privk, token ^ nonce) || H(ephemeral-pubk) || pubk || nonce || 0x1)

        :param remote_pubkey: public key of remote node
        :param nonce: nonce
        :return: authentication message bytes
        """

        assert self._is_initiator

        ecdh_shared_secret = self._ecc.get_ecdh_key(self._remote_pubkey)
        token = ecdh_shared_secret
        flag = 0x0
        self._initiator_nonce = crypto_utils.keccak_hash(rlp_utils.int_to_big_endian(random.randint(0, sys.maxsize)))
        assert len(self._initiator_nonce) == eth_constants.SHA3_LEN_BYTES

        token_xor_nonce = crypto_utils.string_xor(token, self._initiator_nonce)
        assert len(token_xor_nonce) == eth_constants.SHA3_LEN_BYTES

        ephemeral_pubkey = self._ephemeral_ecc.get_raw_public_key()
        assert len(ephemeral_pubkey) == eth_constants.PUBLIC_KEY_LEN
        if not self._ecc.is_valid_key(ephemeral_pubkey):
            raise InvalidKeyError("Invalid ephemeral pubkey")

        # S(ephemeral-privk, ecdh-shared-secret ^ nonce)
        S = self._ephemeral_ecc.sign(token_xor_nonce)
        assert len(S) == eth_constants.SIGNATURE_LEN

        # S || H(ephemeral-pubk) || pubk || nonce || 0x0
        auth_message = S + crypto_utils.keccak_hash(ephemeral_pubkey) + self._ecc.get_raw_public_key() + \
                       self._initiator_nonce + rlp_utils.ascii_chr(flag)
        return auth_message

    def encrypt_auth_message(self, auth_message):
        """
        Encrypts authentication message
        :param auth_message: message bytes
        :param remote_pubkey: remote public key
        :return: encrypted message
        """

        assert self._is_initiator

        self._auth_init = self._ecc.encrypt(auth_message, self._remote_pubkey)
        return self._auth_init

    def decrypt_auth_message(self, cipher_text):
        """
        Decrypts authentication message
        :param cipher_text: encrypted message
        :return: decrypted authentication message
        """

        assert not self._is_initiator

        if len(cipher_text) < eth_constants.ENC_AUTH_MSG_LEN:
            logger.debug("Minimal len of auth message is {0} but was {1}. Continue waiting for more bytes."
                         .format(eth_constants.ENC_AUTH_MSG_LEN, len(cipher_text)))
            return None, None

        try:
            if len(cipher_text) == eth_constants.ENC_AUTH_MSG_LEN:
                # Decrypt plain auth message
                size = eth_constants.ENC_AUTH_MSG_LEN
                message = self._ecc.decrypt(cipher_text[:size])
            else:
                # Decrypt EIP8 auth message
                size = struct.unpack(">H", cipher_text[:eth_constants.EIP8_AUTH_PREFIX_LEN])[0] + \
                       eth_constants.EIP8_AUTH_PREFIX_LEN

                if len(cipher_text) < size:
                    logger.debug("Expected auth message size is {0} but was {1}. Continue waiting more bytes."
                                 .format(size, len(cipher_text)))
                    return None, None

                message = self._ecc.decrypt(bytes(cipher_text[eth_constants.EIP8_AUTH_PREFIX_LEN:size]),
                                            shared_mac_data=bytes(cipher_text[:eth_constants.EIP8_AUTH_PREFIX_LEN]))
                self._is_eip8_auth = True
        except RuntimeError as e:
            raise AuthenticationError(e)

        self._auth_init = cipher_text[:size]
        return message, size

    def parse_auth_message(self, message):
        assert not self._is_initiator

        if self._is_eip8_auth:
            (signature, pubkey, nonce, version) = self.parse_eip8_auth_message(message)
        else:
            (signature, pubkey, nonce, version) = self.parse_plain_auth_message(message)

        token = self._ecc.get_ecdh_key(pubkey)
        remote_ephemeral_pubkey = crypto_utils.recover_public_key(crypto_utils.string_xor(token, nonce),
                                                                  signature)
        if not self._ecc.is_valid_key(remote_ephemeral_pubkey):
            raise InvalidKeyError("Invalid remote ephemeral pubkey")

        self._remote_ephemeral_pubkey = remote_ephemeral_pubkey
        self._initiator_nonce = nonce
        self._remote_pubkey = pubkey
        self._remote_version = eth_constants.P2P_PROTOCOL_VERSION

    def parse_plain_auth_message(self, message):
        if len(message) != eth_constants.AUTH_MSG_LEN:
            raise ValueError("Authentication message of len {0} is expected but was {1}"
                             .format(eth_constants.ENC_AUTH_MSG_LEN, len(message)))

        signature = message[:eth_constants.SIGNATURE_LEN]

        pubkey_start = eth_constants.SIGNATURE_LEN + eth_constants.MAC_LEN
        pubkey = message[pubkey_start:pubkey_start + eth_constants.PUBLIC_KEY_LEN]
        if not self._ecc.is_valid_key(pubkey):
            raise InvalidKeyError("Invalid initiator pubkey")

        nonce_start = eth_constants.SIGNATURE_LEN + eth_constants.MAC_LEN + eth_constants.PUBLIC_KEY_LEN
        nonce = message[nonce_start: nonce_start + eth_constants.AUTH_NONCE_LEN]
        known_flag = bool(rlp_utils.safe_ord(message[nonce_start + eth_constants.AUTH_NONCE_LEN:]))
        assert known_flag == 0

        return signature, pubkey, nonce, eth_constants.P2P_PROTOCOL_VERSION

    def parse_eip8_auth_message(self, message):
        eip8_auth_serializers = sedes.List(
            [
                sedes.Binary(min_length=eth_constants.SIGNATURE_LEN, max_length=eth_constants.SIGNATURE_LEN),  # sig
                sedes.Binary(min_length=eth_constants.PUBLIC_KEY_LEN, max_length=eth_constants.PUBLIC_KEY_LEN),
                # pubkey
                sedes.Binary(min_length=eth_constants.AUTH_NONCE_LEN, max_length=eth_constants.AUTH_NONCE_LEN),  # nonce
                sedes.BigEndianInt()  # version
            ], strict=False)

        values = rlp.decode(message, sedes=eip8_auth_serializers, strict=False)
        signature, pubkey, nonce, version = values

        return signature, pubkey, nonce, version

    def create_auth_ack_message(self):
        """
        authRecipient = E(remote-pubk, remote-ephemeral-pubk || nonce || 0x1) // token found
        authRecipient = E(remote-pubk, remote-ephemeral-pubk || nonce || 0x0) // token not found

        nonce, ephemeral_pubkey, version are local!
        """

        assert not self._is_initiator

        ephemeral_pubkey = self._ephemeral_ecc.get_raw_public_key()
        encoded_nonce = rlp.sedes.big_endian_int.serialize(random.randint(0, eth_constants.MAX_NONCE))
        self._responder_nonce = crypto_utils.keccak_hash(encoded_nonce)

        if self._is_eip8_auth:
            msg = self.create_eip8_auth_ack_message(ephemeral_pubkey, self._responder_nonce,
                                                    eth_constants.P2P_PROTOCOL_VERSION)
            assert len(msg) > eth_constants.AUTH_ACK_MSG_LEN
        else:
            msg = self.create_plain_auth_ack_message(ephemeral_pubkey, self._responder_nonce,
                                                     eth_constants.P2P_PROTOCOL_VERSION)
            assert len(msg) == eth_constants.AUTH_ACK_MSG_LEN

        return msg

    def create_plain_auth_ack_message(self, ephemeral_pubkey, nonce, version):
        return ephemeral_pubkey + nonce + b"\x00"


    def create_eip8_auth_ack_message(self, ephemeral_pubkey, nonce, version):
        eip8_ack_serializers = sedes.List(
            [
                sedes.Binary(min_length=eth_constants.PUBLIC_KEY_LEN, max_length=eth_constants.PUBLIC_KEY_LEN),  # ephemeral pubkey
                sedes.Binary(min_length=eth_constants.AUTH_NONCE_LEN, max_length=eth_constants.AUTH_NONCE_LEN),  # nonce
                sedes.BigEndianInt()  # version
            ], strict=False)

        data = rlp.encode((ephemeral_pubkey, nonce, version), sedes=eip8_ack_serializers)
        pad = os.urandom(random.randint(eth_constants.EIP8_ACK_PAD_MIN, eth_constants.EIP8_ACK_PAD_MAX))
        return data + pad

    def encrypt_auth_ack_message(self, ack_message):
        assert not self._is_initiator

        if self._is_eip8_auth:
            prefix = struct.pack(">H", len(ack_message) + eth_constants.ECIES_ENCRYPT_OVERHEAD_LENGTH)
            enc_auth_ack = self._ecc.encrypt(ack_message, self._remote_pubkey, shared_mac_data=prefix)

            self._auth_ack = prefix + enc_auth_ack
        else:
            self._auth_ack = self._ecc.encrypt(ack_message, self._remote_pubkey)
            assert len(self._auth_ack) == eth_constants.ENC_AUTH_ACK_MSG_LEN

        return self._auth_ack

    def decrypt_auth_ack_message(self, cipher_text):
        """
        Decrypts authentication acknowledge message
        :param cipher_text: encrypted message
        :return: decrypted message
        """

        assert self._is_initiator

        if len(cipher_text) != eth_constants.ENC_AUTH_ACK_MSG_LEN:
            raise ValueError("Expected length of auth ack message is {0} but was provided {1}"
                             .format(eth_constants.ENC_AUTH_ACK_MSG_LEN, len(cipher_text)))

        (size, eph_pubkey, nonce) = self._decode_auth_ack_message(cipher_text)
        self._auth_ack = cipher_text[:size]
        self._remote_ephemeral_pubkey = eph_pubkey[:eth_constants.PUBLIC_KEY_LEN]
        self._responder_nonce = nonce
        self._remote_version = eth_constants.AUTH_MSG_VERSION

        if not self._ecc.is_valid_key(self._remote_ephemeral_pubkey):
            raise InvalidKeyError("Invalid remote ephemeral pubkey")

        return cipher_text[size:]

    def setup_cipher(self):
        """
        Sets up cipher parameters.
        Needs to be called after initial authentication handshake
        """

        assert self._responder_nonce
        assert self._initiator_nonce
        assert self._auth_init
        assert self._auth_ack
        assert self._remote_ephemeral_pubkey
        if not self._ecc.is_valid_key(self._remote_ephemeral_pubkey):
            raise InvalidKeyError("Invalid remote ephemeral pubkey")

        # derive base secrets from ephemeral key agreement
        # ecdhe-shared-secret = ecdh.agree(ephemeral-privkey, remote-ephemeral-pubk)
        ecdhe_shared_secret = self._ephemeral_ecc.get_ecdh_key(self._remote_ephemeral_pubkey)

        # shared-secret = sha3(ecdhe-shared-secret || sha3(nonce || initiator-nonce))
        shared_secret = crypto_utils.keccak_hash(
            ecdhe_shared_secret + crypto_utils.keccak_hash(self._responder_nonce + self._initiator_nonce))

        self.ecdhe_shared_secret = ecdhe_shared_secret  # used in tests
        self.shared_secret = shared_secret  # used in tests

        # token = sha3(shared-secret)
        self._token = crypto_utils.keccak_hash(shared_secret)

        # aes-secret = sha3(ecdhe-shared-secret || shared-secret)
        self._aes_secret = crypto_utils.keccak_hash(ecdhe_shared_secret + shared_secret)

        # mac-secret = sha3(ecdhe-shared-secret || aes-secret)
        self.mac_secret = crypto_utils.keccak_hash(ecdhe_shared_secret + self._aes_secret)

        # setup sha3 instances for the MACs
        # egress-mac = sha3.update(mac-secret ^ recipient-nonce || auth-sent-init)
        mac1 = crypto_utils.get_sha3_calculator(
            crypto_utils.string_xor(self.mac_secret, self._responder_nonce) + self._auth_init)
        # ingress-mac = sha3.update(mac-secret ^ initiator-nonce || auth-recvd-ack)
        mac2 = crypto_utils.get_sha3_calculator(
            crypto_utils.string_xor(self.mac_secret, self._initiator_nonce) + self._auth_ack)

        if self._is_initiator:
            self._egress_mac, self._ingress_mac = mac1, mac2
        else:
            self._egress_mac, self._ingress_mac = mac2, mac1

        iv = "\x00" * eth_constants.IV_LEN
        self._aes_enc = pyelliptic.Cipher(self._aes_secret, iv, eth_constants.CIPHER_ENCRYPT_DO,
                                          ciphername=eth_constants.RLPX_CIPHER_NAME)
        self._aes_dec = pyelliptic.Cipher(self._aes_secret, iv, eth_constants.CIPHER_DECRYPT_DO,
                                          ciphername=eth_constants.RLPX_CIPHER_NAME)
        self._mac_enc = AES.new(self.mac_secret, AES.MODE_ECB).encrypt

        self._is_ready = True

    def encrypt_frame(self, frame):
        """
        Encrypts frame
        :param frame: frame data
        :return: encrypted frame
        """

        if not isinstance(frame, Frame):
            raise TypeError("frame must be of type Frame but was {0}".format(type(frame)))

        if not self._is_ready:
            raise CipherNotInitializedError(f"failed to encrypt frame {frame}, the cipher was never initialized!")

        header = frame.get_header()
        body = frame.get_body()

        # header
        header_ciphertext = self.aes_encode(header)

        # egress-mac.update(aes(mac-secret,egress-mac) ^ header-ciphertext).digest
        header_mac = self.mac_egress(
            crypto_utils.string_xor(self._mac_enc(self.mac_egress()[:eth_constants.FRAME_MAC_LEN]),
                                    header_ciphertext))[:eth_constants.FRAME_MAC_LEN]

        # frame
        frame_ciphertext = self.aes_encode(body)
        assert len(frame_ciphertext) == len(body)
        # egress-mac.update(aes(mac-secret,egress-mac) ^
        # left128(egress-mac.update(frame-ciphertext).digest))
        fmac_seed = self.mac_egress(frame_ciphertext)
        frame_mac = self.mac_egress(
            crypto_utils.string_xor(self._mac_enc(self.mac_egress()[:eth_constants.FRAME_MAC_LEN]),
                                    fmac_seed[:eth_constants.FRAME_MAC_LEN]))[:eth_constants.FRAME_MAC_LEN]

        return bytearray(header_ciphertext + header_mac + frame_ciphertext + frame_mac)

    def decrypt_frame_header(self, data):
        """
        Decrypts frame header
        :param data: frame header data
        :return: decrypted header data
        """

        if len(data) != eth_constants.FRAME_HDR_TOTAL_LEN:
            raise ValueError("Frame header len is expected to be {0} but was {1}"
                             .format(eth_constants.FRAME_HDR_TOTAL_LEN, len(data)))

        if not self._is_ready:
            raise CipherNotInitializedError("failed to decrypt frame header, the cipher was never initialized!")

        header_cipher_text = data[:eth_constants.FRAME_HDR_DATA_LEN]
        header_mac = data[eth_constants.FRAME_HDR_DATA_LEN:eth_constants.FRAME_HDR_TOTAL_LEN]

        # ingress-mac.update(aes(mac-secret,ingress-mac) ^ header-ciphertext).digest
        mac1 = self.mac_ingress()[:eth_constants.FRAME_MAC_LEN]
        mac_enc = self._mac_enc(mac1)
        mac_sxor = crypto_utils.string_xor(mac_enc, header_cipher_text)

        expected_header_mac = self.mac_ingress(mac_sxor)[:eth_constants.FRAME_MAC_LEN]

        # expected_header_mac = self.updateMAC(self.ingress_mac, header_ciphertext)
        if not expected_header_mac == header_mac:
            raise AuthenticationError("Invalid header mac")

        return self.aes_decode(header_cipher_text)

    def decrypt_frame_body(self, data, body_size):
        """
        Decrypts frame body
        :param data: frame data
        :param body_size: body size
        :return: decrypted frame body
        """

        if not self._is_ready:
            raise CipherNotInitializedError("failed to decrypt frame body, the cipher was never initialized!")


        # frame-size: 3-byte integer size of frame, big endian encoded (excludes padding)
        # frame relates to body w/o padding w/o mac

        read_size = crypto_utils.get_padded_len_16(body_size)

        if not len(data) >= read_size + eth_constants.FRAME_MAC_LEN:
            raise ParseError("Insufficient body length")

        frame_cipher_text = data[:read_size]
        frame_mac = data[read_size:read_size + eth_constants.FRAME_MAC_LEN]

        # ingres-mac.update(aes(mac-secret,ingres-mac) ^
        # left128(ingres-mac.update(frame-ciphertext).digest))
        frame_mac_seed = self.mac_ingress(frame_cipher_text)
        expected_frame_mac = self.mac_ingress(
            crypto_utils.string_xor(self._mac_enc(self.mac_ingress()[:eth_constants.FRAME_MAC_LEN]),
                                    frame_mac_seed[:eth_constants.FRAME_MAC_LEN]))[:eth_constants.FRAME_MAC_LEN]

        if not frame_mac == expected_frame_mac:
            raise AuthenticationError("Invalid frame mac")

        return self.aes_decode(frame_cipher_text)[:body_size]

    def aes_encode(self, data=""):
        if isinstance(data, bytearray):
            bytes_to_encode = bytes(data)
        elif isinstance(data, memoryview):
            bytes_to_encode = data.tobytes()
        else:
            bytes_to_encode = data

        return self._aes_enc.update(bytes_to_encode)

    def aes_decode(self, data=""):
        if isinstance(data, bytearray):
            bytes_to_decode = bytes(data)
        elif isinstance(data, memoryview):
            bytes_to_decode = data.tobytes()
        else:
            bytes_to_decode = data

        return self._aes_dec.update(bytes_to_decode)

    def mac_egress(self, data=b""):
        data = rlp_utils.str_to_bytes(data)
        self._egress_mac.update(data)
        return self._egress_mac.digest()

    def mac_ingress(self, data=b""):
        data = rlp_utils.str_to_bytes(data)
        self._ingress_mac.update(data)
        return self._ingress_mac.digest()

    def _decode_auth_ack_message(self, cipher_text):
        try:
            message = self._ecc.decrypt(cipher_text[:eth_constants.ENC_AUTH_ACK_MSG_LEN])
        except RuntimeError as e:
            raise AuthenticationError(e)
        eph_pubkey = message[:eth_constants.PUBLIC_KEY_LEN]
        nonce = message[eth_constants.PUBLIC_KEY_LEN:eth_constants.PUBLIC_KEY_LEN + eth_constants.AUTH_NONCE_LEN]
        known = rlp_utils.safe_ord(message[-1])
        assert known == 0
        return (eth_constants.ENC_AUTH_ACK_MSG_LEN, eph_pubkey, nonce)



from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxgateway.feed.eth.eth_new_transaction_feed import EthNewTransactionFeed
from bxgateway.feed.eth.eth_raw_transaction import EthRawTransaction


class EthRelayConnection(AbstractRelayConnection):
    def publish_new_transaction(
        self, tx_hash: Sha256Hash, tx_contents: memoryview
    ) -> None:
        self.node.feed_manager.publish_to_feed(
            EthNewTransactionFeed.NAME,
            EthRawTransaction(tx_hash, tx_contents)
        )





import argparse
import sys

from bxgateway import eth_constants
from bxutils import logging


logger = logging.get_logger(__name__)


class ParseEnode(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        assert isinstance(values, str)
        enode = values

        # Make sure enode is at least as long as the public key
        if len(enode) < 2 * eth_constants.PUBLIC_KEY_LEN:
            logger.fatal("Invalid enode. "
                         "Invalid enode length: {}", len(enode), exc_info=False)
            sys.exit(1)
        try:
            enode_and_pub_key, ip_and_port = enode.split("@")
            if enode_and_pub_key.startswith("enode://"):
                pub_key = enode_and_pub_key[8:]
            else:
                pub_key = enode_and_pub_key
            ip, port_and_disc = ip_and_port.split(":")
            port = port_and_disc.split("?")[0]
        except ValueError:
            logger.fatal("Invalid enode: {}", enode, exc_info=False)
            sys.exit(1)
        else:
            # Node public key gets validated in validate_eth_opts
            namespace.node_public_key = pub_key
            # blockchain IP gets validated in __init__()
            namespace.blockchain_ip = ip
            # Port validation
            if not port.isnumeric():
                logger.fatal("Invalid port: {}", port, exc_info=False)
                sys.exit(1)
            namespace.blockchain_port = int(port)



import typing
from typing import Union, cast, List, NamedTuple

from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.abstract_message_converter import AbstractMessageConverter
from bxgateway.messages.btc.tx_btc_message import TxBtcMessage
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import TransactionsEthProtocolMessage


class ProcessTransactionMessageFromNodeResult(NamedTuple):
    seen: bool
    transaction_hash: Sha256Hash
    transaction_contents: Union[bytearray, memoryview]
    bdn_transaction_message: TxMessage


class OntTxMessage(object):
    pass


class GatewayTransactionService(TransactionService):

    def process_transactions_message_from_node(
        self,
        msg: Union[TxBtcMessage, TransactionsEthProtocolMessage, OntTxMessage]
    ) -> List[ProcessTransactionMessageFromNodeResult]:

        # pyre-fixme[16]: `bxcommon.connections.abstract_node.AbstractNode` has no attribute `message_converter`.
        message_converter = cast(AbstractMessageConverter, self.node.message_converter)

        # pyre-fixme[16]: `bxcommon.connections.abstract_node.AbstractNode` has no attribute `default_tx_quota_type`.
        bx_tx_messages = message_converter.tx_to_bx_txs(msg, self.network_num, self.node.default_tx_quota_type)

        result = []

        for bx_tx_message, tx_hash, tx_bytes in bx_tx_messages:

            tx_cache_key = self._tx_hash_to_cache_key(tx_hash)

            tx_seen_flag = self.has_transaction_contents_by_cache_key(tx_cache_key) or \
                           self.removed_transaction_by_cache_key(tx_cache_key)

            if not tx_seen_flag:
                self.set_transaction_contents(tx_hash, tx_bytes)

            result.append(ProcessTransactionMessageFromNodeResult(
                tx_seen_flag,
                tx_hash,
                tx_bytes,
                bx_tx_message
            ))

        return result



import datetime
import time
from typing import Tuple, Optional, List, cast

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.models.quota_type_model import QuotaType
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.abstract_message_converter import AbstractMessageConverter, BlockDecompressionResult
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.utils.block_info import BlockInfo


class BtcNoCompressionMessageConverter(AbstractMessageConverter):
    def tx_to_bx_txs(self, tx_msg, network_num, quota_type: Optional[QuotaType] = None):
        raise NotImplementedError()

    def bx_tx_to_tx(self, bx_tx_msg):
        raise NotImplementedError()

    def block_to_bx_block(self, block_msg, tx_service) -> Tuple[memoryview, BlockInfo]:
        start_datetime = datetime.datetime.utcnow()
        start_time = time.time()

        block_msg = cast(BlockBtcMessage, block_msg)

        block_info = BlockInfo(
            block_msg.block_hash(),
            [],
            start_datetime,
            datetime.datetime.utcnow(),
            (time.time() - start_time) * 1000,
            block_msg.txn_count(),
            # pyre-fixme[6]: Expected `Optional[str]` for 7th param but got
            #  `BtcObjectHash`.
            block_msg.block_hash(),
            convert.bytes_to_hex(block_msg.prev_block_hash().binary),
            len(block_msg.rawbytes()),
            len(block_msg.rawbytes()),
            0
        )
        return block_msg.rawbytes(), block_info

    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        start_datetime = datetime.datetime.utcnow()
        start_time = time.time()
        block_msg = BlockBtcMessage(buf=bx_block_msg)

        block_info = BlockInfo(
            block_msg.block_hash(),
            [],
            start_datetime,
            datetime.datetime.utcnow(),
            (time.time() - start_time) * 1000,
            block_msg.txn_count(),
            # pyre-fixme[6]: Expected `Optional[str]` for 7th param but got
            #  `BtcObjectHash`.
            block_msg.block_hash(),
            convert.bytes_to_hex(block_msg.prev_block_hash().binary),
            len(block_msg.rawbytes()),
            len(block_msg.rawbytes()),
            0
        )
        return BlockDecompressionResult(block_msg, block_info, [], [])



from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.v8.broadcast_message_converter_v8 import broadcast_message_converter_v8
from bxcommon.messages.bloxroute.v9.bdn_performance_stats_message_converter_v9 \
    import bdn_performance_stats_message_converter_v9
from bxcommon.messages.versioning.abstract_version_converter_factory import AbstractMessageConverterFactory
from bxcommon.messages.versioning.no_changes_message_converter import no_changes_message_converter


class _MessageConverterFactoryV8(AbstractMessageConverterFactory):
    _MESSAGE_CONVERTER_MAPPING = {
        BloxrouteMessageType.BDN_PERFORMANCE_STATS: bdn_performance_stats_message_converter_v9,
        BloxrouteMessageType.BROADCAST: broadcast_message_converter_v8,
    }

    def get_message_converter(self, msg_type):
        if not msg_type:
            raise ValueError("msg_type is required.")

        if msg_type not in self._MESSAGE_CONVERTER_MAPPING:
            return no_changes_message_converter

        return self._MESSAGE_CONVERTER_MAPPING[msg_type]


message_converter_factory_v8 = _MessageConverterFactoryV8()



import datetime
from typing import Optional

from bxcommon.utils.stats.stat_event_type_settings import StatEventTypeSettings


class StatEvent:
    def __init__(self, event_settings: StatEventTypeSettings, event_subject_id: str, node_id: str,
                 start_date_time: datetime.datetime, end_date_time=Optional[datetime.datetime], **kwargs):
        self.event_name = event_settings.name
        self.event_logic = event_settings.event_logic
        self.event_subject_id = event_subject_id
        self.node_id = node_id
        self.start_date_time = start_date_time
        self.end_date_time = end_date_time if end_date_time is not None else start_date_time

        self.extra_data = kwargs



import gc
import time
from datetime import datetime
from typing import Dict, Any, Optional

from bxcommon import constants
from bxcommon.utils.stats.node_statistics_service import node_stats_service
from bxutils import logging
from bxutils.logging import LogRecordType

logger = logging.get_logger(LogRecordType.GarbageCollection)

_gc_start: Optional[float] = None


def gc_callback(phase: str, info: Dict[str, Any]):
    # pylint: disable=global-statement
    global _gc_start
    gc_start = _gc_start

    if phase == "start" and gc_start is None:
        _gc_start = time.time()
    elif gc_start is not None:
        duration = time.time() - gc_start
        _gc_start = None

        if node_stats_service.node is not None:
            node_stats_service.log_gc_duration(info["generation"], duration)
        gen0, gen1, gen2 = gc.get_count()
        if duration >= constants.GC_DURATION_WARN_THRESHOLD:
            logger.statistics(
                {
                    "type": "GcDurationExceededWarningThreshold",
                    "start_time": datetime.fromtimestamp(time.time()),
                    "duration": duration,
                    "generation": info["generation"],
                    "collected": info["collected"],
                    "uncollectable": info["uncollectable"],
                    "total_uncollectable": len(gc.garbage),
                    "sizes": {"generation0": gen0, "generation1": gen1, "generation2": gen2},
                }
            )
    else:
        logger.debug("invalid state when attempting to track GC state skip")



import rlp

from bxgateway.messages.eth.serializers.block_header import BlockHeader
from bxgateway.messages.eth.serializers.transaction import Transaction


class TransientBlockBody(rlp.Serializable):
    fields = [
        ("transactions", rlp.sedes.CountableList(Transaction)),
        ("uncles", rlp.sedes.CountableList(BlockHeader))
    ]




from typing import Tuple, Optional, Union, List

from bxutils import logging

from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.models.quota_type_model import QuotaType
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.blockchain_utils.bdn_tx_to_bx_tx import bdn_tx_to_bx_tx
from bxcommon.utils.blockchain_utils.eth.eth_common_util import raw_tx_to_bx_tx
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.abstract_message_converter import AbstractMessageConverter, BlockDecompressionResult
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import TransactionsEthProtocolMessage
from bxgateway.utils.block_info import BlockInfo
from bxgateway.utils.eth import rlp_utils


logger = logging.get_logger(__name__)


def parse_transaction_bytes(tx_bytes: memoryview) -> TransactionsEthProtocolMessage:
    size = len(tx_bytes)

    txs_prefix = rlp_utils.get_length_prefix_list(size)
    size += len(txs_prefix)

    buf = bytearray(size)

    buf[0:len(txs_prefix)] = txs_prefix
    buf[len(txs_prefix):] = tx_bytes
    return TransactionsEthProtocolMessage(buf)


class EthAbstractMessageConverter(AbstractMessageConverter):

    def __init__(self):
        self._last_recovery_idx: int = 0

    def tx_to_bx_txs(self, tx_msg, network_num, quota_type: Optional[QuotaType] = None) ->\
            List[Tuple[TxMessage, Sha256Hash, Union[bytearray, memoryview]]]:
        """
        Converts Ethereum transactions message to array of internal transaction messages

        The code is optimized and does not make copies of bytes

        :param tx_msg: Ethereum transaction message
        :param network_num: blockchain network number
        :param quota_type: the quota type to assign to the BDN transaction.
        :return: array of tuples (transaction message, transaction hash, transaction bytes)
        """

        if not isinstance(tx_msg, TransactionsEthProtocolMessage):
            raise TypeError("TransactionsEthProtocolMessage is expected for arg tx_msg but was {0}"
                            .format(type(tx_msg)))
        bx_tx_msgs = []

        msg_bytes = memoryview(tx_msg.rawbytes())

        _, length, start = rlp_utils.consume_length_prefix(msg_bytes, 0)
        txs_bytes = msg_bytes[start:]

        tx_start_index = 0

        while True:
            bx_tx, tx_item_length, tx_item_start = raw_tx_to_bx_tx(txs_bytes, tx_start_index, network_num, quota_type)
            bx_tx_msgs.append((bx_tx, bx_tx.message_hash(), bx_tx.tx_val()))

            tx_start_index = tx_item_start + tx_item_length

            if tx_start_index == len(txs_bytes):
                break

        return bx_tx_msgs

    def bdn_tx_to_bx_tx(
            self,
            raw_tx: Union[bytes, bytearray, memoryview],
            network_num: int,
            quota_type: Optional[QuotaType] = None
    ) -> TxMessage:
        return bdn_tx_to_bx_tx(raw_tx, network_num, quota_type)

    def bx_tx_to_tx(self, bx_tx_msg):
        """
        Converts internal transaction message to Ethereum transactions message

        The code is optimized and does not make copies of bytes

        :param bx_tx_msg: internal transaction message
        :return: Ethereum transactions message
        """

        if not isinstance(bx_tx_msg, TxMessage):
            raise TypeError("Type TxMessage is expected for bx_tx_msg arg but was {0}"
                            .format(type(bx_tx_msg)))

        return parse_transaction_bytes(bx_tx_msg.tx_val())

    def block_to_bx_block(self, block_msg: InternalEthBlockInfo, tx_service) -> Tuple[
        memoryview, BlockInfo]:
        """
        Convert Ethereum new block message to internal broadcast message with transactions replaced with short ids

        The code is optimized and does not make copies of bytes

        :param block_msg: Ethereum new block message
        :param tx_service: Transactions service
        :return: Internal broadcast message bytes (bytearray), tuple (txs count, previous block hash)
        """
        raise NotImplementedError

    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        """
        Converts internal broadcast message to Ethereum new block message

        The code is optimized and does not make copies of bytes

        :param bx_block_msg: internal broadcast message bytes
        :param tx_service: Transactions service
        :return: tuple (new block message, block hash, unknown transaction short id, unknown transaction hashes)
        """
        raise NotImplementedError



from collections import defaultdict
from typing import TYPE_CHECKING, Dict, Set, List, Optional, Iterator, cast, Tuple

from bxcommon import constants
from bxcommon.utils import memory_utils, crypto
from bxcommon.utils.alarm_queue import AlarmId
from bxcommon.utils.expiring_dict import ExpiringDict
from bxcommon.utils.memory_utils import ObjectSize
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import hooks
from bxgateway import eth_constants, gateway_constants
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxgateway.messages.eth.new_block_parts import NewBlockParts
from bxgateway.messages.eth.protocol.block_bodies_eth_protocol_message import (
    BlockBodiesEthProtocolMessage,
)
from bxgateway.messages.eth.protocol.block_headers_eth_protocol_message import (
    BlockHeadersEthProtocolMessage,
)
from bxgateway.messages.eth.protocol.get_block_headers_eth_protocol_message import (
    GetBlockHeadersEthProtocolMessage,
)
from bxgateway.messages.eth.protocol.new_block_eth_protocol_message import (
    NewBlockEthProtocolMessage,
)
from bxgateway.messages.eth.protocol.new_block_hashes_eth_protocol_message import (
    NewBlockHashesEthProtocolMessage,
)
from bxgateway.services.push_block_queuing_service import (
    PushBlockQueuingService,
)
from bxutils import logging

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(__name__)


class EthBlockQueuingService(
    PushBlockQueuingService[
        InternalEthBlockInfo, BlockHeadersEthProtocolMessage
    ]
):
    block_checking_alarms: Dict[Sha256Hash, AlarmId]
    block_repeat_count: Dict[Sha256Hash, int]
    _block_parts: ExpiringDict[Sha256Hash, NewBlockParts]
    _block_hashes_by_height: ExpiringDict[int, Set[Sha256Hash]]
    _height_by_block_hash: ExpiringDict[Sha256Hash, int]
    _highest_block_number: int = 0

    def __init__(self, node: "AbstractGatewayNode"):
        super().__init__(node)
        self.node: "EthGatewayNode" = cast("EthGatewayNode", node)
        self.block_checking_alarms = {}
        self.block_repeat_count = defaultdict(int)
        self._block_parts = ExpiringDict(
            node.alarm_queue,
            gateway_constants.MAX_BLOCK_CACHE_TIME_S,
            "eth_block_queue_parts",
        )
        self._block_hashes_by_height = ExpiringDict(
            node.alarm_queue,
            gateway_constants.MAX_BLOCK_CACHE_TIME_S,
            "eth_block_queue_hashes_by_heights",
        )
        self._height_by_block_hash = ExpiringDict(
            node.alarm_queue,
            gateway_constants.MAX_BLOCK_CACHE_TIME_S,
            "eth_block_queue_height_by_hash"
        )

    def build_block_header_message(
        self, block_hash: Sha256Hash, block_message: InternalEthBlockInfo
    ) -> BlockHeadersEthProtocolMessage:
        if block_hash in self._block_parts:
            block_header_bytes = self._block_parts[
                block_hash
            ].block_header_bytes
        else:
            block_header_bytes = (
                block_message.to_new_block_parts().block_header_bytes
            )
        return BlockHeadersEthProtocolMessage.from_header_bytes(
            block_header_bytes
        )

    def send_block_to_node(
        self, block_hash: Sha256Hash, block_msg: Optional[InternalEthBlockInfo],
    ) -> None:
        assert block_msg is not None
        new_block_parts = self._block_parts[block_hash]

        if block_msg.has_total_difficulty():
            new_block_msg = block_msg.to_new_block_msg()
            super(EthBlockQueuingService, self).send_block_to_node(
                block_hash, new_block_msg
            )
            self.node.set_known_total_difficulty(
                new_block_msg.block_hash(), new_block_msg.chain_difficulty()
            )
        else:
            calculated_total_difficulty = self.node.try_calculate_total_difficulty(
                block_hash, new_block_parts
            )

            if calculated_total_difficulty is None:
                # Total difficulty may be unknown after a long fork or
                # if gateways just started. Announcing new block hashes to
                # ETH node in that case. It
                # will request header and body separately.
                new_block_headers_msg = NewBlockHashesEthProtocolMessage.from_block_hash_number_pair(
                    block_hash, new_block_parts.block_number
                )
                super(EthBlockQueuingService, self).send_block_to_node(
                    block_hash, new_block_headers_msg
                )
            else:
                new_block_msg = NewBlockEthProtocolMessage.from_new_block_parts(
                    new_block_parts, calculated_total_difficulty
                )
                super(EthBlockQueuingService, self).send_block_to_node(
                    block_hash, new_block_msg
                )

        self.node.log_blocks_network_content(self.node.network_num, block_msg)

    # pyre-fixme[14]: `get_previous_block_hash_from_message` overrides method defined in
    #  `PushBlockQueuingService` inconsistently.
    def get_previous_block_hash_from_message(
        self, block_message: NewBlockEthProtocolMessage
    ) -> Sha256Hash:
        return block_message.prev_block_hash()

    def get_block_parts(self, block_hash: Sha256Hash) -> Optional[NewBlockParts]:
        if block_hash not in self._block_parts:
            logger.debug("requested transaction info for a block not in the queueing service {}", block_hash)
            return None
        return self._block_parts[block_hash]

    def get_block_body_from_message(self, block_hash: Sha256Hash) -> Optional[BlockBodiesEthProtocolMessage]:
        block_parts = self.get_block_parts(block_hash)
        if block_parts is None:
            return None
        return BlockBodiesEthProtocolMessage.from_body_bytes(block_parts.block_body_bytes)

    # pyre-fixme[14]: `on_block_sent` overrides method defined in
    #  `PushBlockQueuingService` inconsistently.
    def on_block_sent(
        self, block_hash: Sha256Hash, _block_message: NewBlockEthProtocolMessage
    ):
        if block_hash not in self.block_checking_alarms:
            # requires delay to have time to validate and process block
            self.block_checking_alarms[
                block_hash
            ] = self.node.alarm_queue.register_alarm(
                eth_constants.CHECK_BLOCK_RECEIPT_DELAY_S,
                self._check_for_block_on_repeat,
                block_hash,
            )

    def push(
        self,
        block_hash: Sha256Hash,
        block_msg: Optional[InternalEthBlockInfo] = None,
        waiting_for_recovery: bool = False,
    ) -> None:
        if not waiting_for_recovery:
            assert block_msg is not None
            self._store_block_parts(block_hash, block_msg)
        super().push(block_hash, block_msg, waiting_for_recovery)

    def store_block_data(
        self,
        block_hash: Sha256Hash,
        block_msg: InternalEthBlockInfo
    ):
        super().store_block_data(block_hash, block_msg)
        self._store_block_parts(block_hash, block_msg)

    def mark_block_seen_by_blockchain_node(
        self,
        block_hash: Sha256Hash,
        block_message: Optional[InternalEthBlockInfo] = None,
    ) -> None:
        if block_message is not None:
            self._store_block_parts(block_hash, block_message)

        super().mark_block_seen_by_blockchain_node(block_hash, block_message)

        if block_hash in self.block_checking_alarms:
            self.node.alarm_queue.unregister_alarm(
                self.block_checking_alarms[block_hash]
            )
            del self.block_checking_alarms[block_hash]
            self.block_repeat_count.pop(block_hash, 0)

    def remove(self, block_hash: Sha256Hash) -> int:
        index = super().remove(block_hash)
        if block_hash in self._block_parts:
            del self._block_parts[block_hash]
            height = self._height_by_block_hash.contents.pop(block_hash, None)
            logger.trace("Removing block {} at height {}", block_hash, height)
            if height:
                self._block_hashes_by_height.contents.get(
                    height, set()
                ).discard(block_hash)
        return index

    def try_send_bodies_to_node(self, block_hashes: List[Sha256Hash]) -> bool:
        """
        Creates and sends block bodies to blockchain connection.
        """
        bodies = []
        for block_hash in block_hashes:
            if block_hash not in self._blocks:
                logger.debug("{} was not found in queue. Aborting.", block_hash)
                return False

            block_message = self._blocks[block_hash]
            if block_message is None:
                logger.debug("{} was not ready in the queue. Aborting", block_hash)
                return False

            if block_hash in self._block_parts:
                block_body_bytes = self._block_parts[block_hash].block_body_bytes
            else:
                block_body_bytes = (
                    block_message.to_new_block_parts().block_body_bytes
                )

            partial_message = BlockBodiesEthProtocolMessage.from_body_bytes(
                block_body_bytes
            )
            block_bodies = partial_message.get_blocks()
            assert len(block_bodies) == 1
            bodies.append(block_bodies[0])

            height = self._height_by_block_hash.contents.get(block_hash, None)
            logger.debug(
                "Appending {} body ({}) for sending to blockchain node.",
                block_hash,
                height
            )

        full_message = BlockBodiesEthProtocolMessage(None, bodies)
        self.node.send_msg_to_node(full_message)
        return True

    def try_send_headers_to_node(self, block_hashes: List[Sha256Hash]) -> bool:
        """
        Creates and sends a block headers message to blockchain connection.

        In most cases, this method should be called with block hashes that are confirmed to
        exist in the block queuing service, but contains checks for safety for otherwise,
        and aborts the function if any headers are not found.
        """
        headers = []
        for block_hash in block_hashes:
            if block_hash not in self._blocks:
                logger.debug("{} was not found in queue. Aborting.", block_hash)
                return False

            block_message = self._blocks[block_hash]
            if block_message is None:
                logger.debug("{} was not ready in the queue. Aborting", block_hash)
                return False

            partial_headers_message = self.build_block_header_message(
                block_hash, block_message
            )
            block_headers = partial_headers_message.get_block_headers()
            assert len(block_headers) == 1
            headers.append(block_headers[0])

            height = self._height_by_block_hash.contents.get(block_hash, None)
            logger.debug(
                "Appending {} header ({}) for sending to blockchain node.",
                block_hash,
                height
            )

        full_header_message = BlockHeadersEthProtocolMessage(None, headers)
        self.node.send_msg_to_node(full_header_message)
        return True

    def get_block_hashes_starting_from_hash(
        self, block_hash: Sha256Hash, max_count: int, skip: int, reverse: bool
    ) -> Tuple[bool, List[Sha256Hash]]:
        """
        Finds up to max_count block hashes in queue that we still have headers
        and block messages queued up for.

        Returns (success, [found_hashes])
        """
        if block_hash not in self._blocks:
            return False, []

        starting_height = self._height_by_block_hash[block_hash]
        logger.trace(
            "Found block {} had height {}. Continuing...",
            block_hash,
            starting_height,
        )
        return self.get_block_hashes_starting_from_height(
            starting_height, max_count, skip, reverse
        )

    def get_block_hashes_starting_from_height(
        self, block_height: int, max_count: int, skip: int, reverse: bool,
    ) -> Tuple[bool, List[Sha256Hash]]:
        """
        Performs a 'best-effort' search for block hashes.

        Gives up if a fork is detected in the requested section of the chain.
        Gives up if a block is requested below the most recent block that's
        not tracked by this service.
        Returns as many blocks as possible if some of the blocks requested
        are in the future and have no been produced yet.

        The resulting list starts at `block_height`, and is ascending if
        reverse=False, and descending if reverse=True.

        Returns (success, [found_hashes])
        """
        block_hashes: List[Sha256Hash] = []
        height = block_height

        if reverse:
            multiplier = -1
        else:
            multiplier = 1

        while (
            len(block_hashes) < max_count
            and height in self._block_hashes_by_height
        ):
            matching_hashes = self._block_hashes_by_height[height]

            # A fork has occurred: give up, and fallback to
            # remote blockchain sync
            if len(matching_hashes) > 1:
                logger.trace(
                    "Detected fork when searching for {} "
                    "block hashes starting from height {}.",
                    max_count,
                    block_height,
                )
                return False, []

            block_hashes.append(
                next(iter(self._block_hashes_by_height[height]))
            )
            height += (1 + skip) * multiplier

        # If a block is requested too far in the past, abort and fallback
        # to remote blockchain sync
        if (
            height < self._highest_block_number
            and height not in self._block_hashes_by_height
            and max_count != len(block_hashes)
        ):
            return False, []

        # Ok, Ethereum expects as many hashes as node contains.
        if max_count != len(block_hashes):
            logger.trace(
                "Could not find all {} requested block hashes. Only got {}.",
                max_count,
                len(block_hashes),
            )

        return True, block_hashes

    def iterate_block_hashes_starting_from_hash(
        self,
        block_hash: Sha256Hash,
        max_count: int = gateway_constants.TRACKED_BLOCK_MAX_HASH_LOOKUP) -> Iterator[Sha256Hash]:
        """
        iterate over cached blocks headers in descending order
        :param block_hash: starting block hash
        :param max_count: max number of elements to return
        :return: Iterator of block hashes in descending order
        """
        block_hash_ = block_hash
        for _ in range(max_count):
            if block_hash_ and block_hash_ in self._block_parts:
                yield block_hash_
                block_hash_ = self._block_parts[block_hash_].get_previous_block_hash()
            else:
                break

    def iterate_recent_block_hashes(
        self,
        max_count: int = gateway_constants.TRACKED_BLOCK_MAX_HASH_LOOKUP) -> Iterator[Sha256Hash]:
        """
        :param max_count:
        :return: Iterator[Sha256Hash] in descending order (last -> first)
        """
        if self._highest_block_number not in self._block_hashes_by_height:
            return iter([])
        block_hashes = self._block_hashes_by_height[self._highest_block_number]
        block_hash = next(iter(block_hashes))
        if len(block_hashes) > 1:
            logger.debug(f"iterating over queued blocks starting for a possible fork {block_hash}")

        return self.iterate_block_hashes_starting_from_hash(block_hash, max_count=max_count)

    def get_block_height(self, block_hash: Sha256Hash) -> Optional[int]:
        block_height: Optional[int] = None
        if block_hash and block_hash in self._height_by_block_hash:
            block_height = self._height_by_block_hash[block_hash]
        return block_height

    def log_memory_stats(self):
        hooks.add_obj_mem_stats(
            self.__class__.__name__,
            self.node.network_num,
            self.block_checking_alarms,
            "block_queue_block_checking_alarms",
            ObjectSize(
                size=len(self.block_checking_alarms) * (crypto.SHA256_HASH_LEN + constants.UL_INT_SIZE_IN_BYTES),
                flat_size=0,
                is_actual_size=False
            ),
            object_item_count=len(self.block_checking_alarms),
            object_type=memory_utils.ObjectType.BASE,
            size_type=memory_utils.SizeType.ESTIMATE
        )

        hooks.add_obj_mem_stats(
            self.__class__.__name__,
            self.node.network_num,
            self.block_repeat_count,
            "block_queue_block_repeat_count",
            ObjectSize(
                size=len(self.block_repeat_count) * (crypto.SHA256_HASH_LEN + constants.UL_INT_SIZE_IN_BYTES),
                flat_size=0,
                is_actual_size=False
            ),
            object_item_count=len(self.block_repeat_count),
            object_type=memory_utils.ObjectType.BASE,
            size_type=memory_utils.SizeType.ESTIMATE
        )

    def _store_block_parts(
        self, block_hash: Sha256Hash, block_message: InternalEthBlockInfo
    ):
        new_block_parts = block_message.to_new_block_parts()
        self._block_parts[block_hash] = new_block_parts
        block_number = block_message.block_number()
        if block_number > 0:
            logger.trace(
                "Adding headers for block {} at height: {}",
                block_hash,
                block_number,
            )
            if block_number in self._block_hashes_by_height:
                self._block_hashes_by_height[block_number].add(block_hash)
            else:
                self._block_hashes_by_height.add(block_number, {block_hash})
            self._height_by_block_hash[block_hash] = block_number
            if block_number > self._highest_block_number:
                self._highest_block_number = block_number
        else:
            logger.trace(
                "No block height could be parsed for block: {}", block_hash
            )

    def _check_for_block_on_repeat(self, block_hash: Sha256Hash) -> float:
        get_confirmation_message = GetBlockHeadersEthProtocolMessage(
            None, block_hash.binary, 1, 0, 0
        )

        self.node.send_msg_to_node(get_confirmation_message)

        if self.block_repeat_count[block_hash] < 5:
            self.block_repeat_count[block_hash] += 1
            return eth_constants.CHECK_BLOCK_RECEIPT_INTERVAL_S
        else:
            del self.block_repeat_count[block_hash]
            del self.block_checking_alarms[block_hash]
            return constants.CANCEL_ALARMS



from typing import Optional
from dataclasses import dataclass
from datetime import datetime, date

from bxcommon.models.bdn_service_model_base import BdnServiceModelBase
from bxcommon import constants


@dataclass
class BdnServiceModelConfigBase:
    msg_quota: Optional[BdnServiceModelBase] = None
    permit: Optional[BdnServiceModelBase] = None
    expire_date: str = constants.EPOCH_DATE

    def is_service_valid(self) -> bool:
        today = datetime.utcnow().date()
        try:
            service_expire_date = date.fromisoformat(self.expire_date)
        except (KeyError, ValueError):
            return False

        return service_expire_date >= today
